/*
 Copyright (C) 2017 SAS Institute Inc. Cary, NC, USA
*/

/**
 \file
   \brief Utility script to delete existing jobflow instances and recreate them. The script can be used to bring a 3.1 installation with multiple 
          Federated Areas (FAs) into a supported (collapsed) single FA state. The script must be executed in the batch mode. Please make sure that prior to 
          executing the script, the designated highest FA contains all required inputs, config sets, base dates, bpmn files, nodes, macros, etc. required 
          for instance re-creation.
   
   \details The script only deletes and recreates instances that have a status of "COMPLETED_SUCCESS" (green checkmark in the IRM WebApp UI). Instances that 
            are already PUBLISHED are NOT processed by the script and left "as-is". For each NON-PUBLISHED instance that belong to either of the statuses
            above, the script gathers information to re-create the instance, then the instance is deleted and a new instance created with the information 
            gathered in the first step. Instances with modified inputs are re-created and re-run after uploading modified input(s). These actions are performed
            by invoking the macro irm_rest_recreate_instances that in turn invokes other macros.

            For a comprehensive overview of the script logic, see macro \link irm_rest_recreate_instances.sas. Prior to the macro invocation, the setup macro
            \link irm_setup.sas is invoked to set LUA paths and update SASAUTOS to include macros contained within source/sas/ucmacros delivered in the 
            script installation directory. 

            Since instances created by a user can be deleted and re-created by the same user, the script must be executed for each user that created
            existing instances. To avoid any conflicts and lock situations, schedule the execution for each user sequentially. Please see the section on 
            Required Parameters below for information how to provide username for script execution.
  
   <b> Required Parameters </b>
   The below required parameters must be initialized with appropriate values before the script is executed:
   1. persistent_area_loc  - Complete path to the IRM Persistent Area (PA). Example: /local/install/IRM/SAS94/PersistentArea
   2. highest_fa_id        - ID of the highest Federated Area (FA) in the system. The FA contains all information for instance re-creation. 
      Re-created instances will be associated with this FA ID. Provide the portion after "fa." Example: If fa.3.1.7, provide 3.1.7
   3. srchostname01        - Host machine name. Example: irmserver.abc.xyz.com
   4. srcirmport01         - Server port. Default: 7980 
   5. srcusername01        - Username credentials of user that created instances. Instances will be deleted and re-created for the user. Example: irmuser1
   6. srcpassword01        - Password credentials for the user above. Can be plain text or SAS-Encoded (will be masked during execution).
   
   <b> Optional Parameters </b>
   The below parameters are provided default values that may be changed (optional):
   1. debug          - Amount of execution information in the log. Default to TRUE. Change to FALSE if DEBUG info is not required in the log.
   2. logOptions     - Amount of macro execution information in the log. Defaulted to blank. Could be changed to mprint mlogic symbolgen and other SAS supported options.
   3. pollInt        - Number of seconds between checks for status of instance creation. Default to 10 seconds.
   4. maxWait        - Maximum seconds to wait for an instance creation to complete. Default to 3600 seconds.

   <b> Script usage </b>
   Invoke SAS in the batch mode to execute the script. An example below. If no log file is specified, a default log file by the same name as the script will be created.
   >/local/install/IRM/sas94/SASHome/SASFoundation/9.4/sas -sysin recreate_flow_instances.sas -log recreate_instances.log

   Apart from the log file, successful script execution also results in the creation of two .html files at the script location. The file names are appended with the specified username.
   1. existing_instances_<username>.html - contains a listing of all instances for the specified username prior to any deletion/re-creation
   2. latest_instances_<username>.html   - contains a listing of all instances for the specified username after script execution
   
   The printing of the below line (without quotes) in the execution log marks completion of script execution:
   "EXITING IRM_REST_RECREATE_INSTANCES"
   
   The execution log should be scanned for any errors (search for ERROR:) and corrective actions taken in case any found.

   <b> Instructions </b>
   0. No user should be logged into IRM during this operation, but IRM must be up
   1. Backup your system (FA, PA, and PostgreSQL database)
   2. If you have any custom code, data, or flow, merge them with the highest priority FA
   3. Run the script (recreate_all_flow_instances.sas)
   4. Delete all but the highest priority FA from metadata using the SMC
   5. Restart IRM (SASServer8)

   \ingroup macroUtils
   
   \author SAS Institute Inc.
   \date 2017
    
 **/

/* Start of script to re-create instances */
/**************************************************************************************
    REQUIRED PARAMETERS WHOSE VALUES MUST BE SET ON A PER SITE BASIS
 **************************************************************************************/
%let persistent_area_loc =; /* Complete path to the IRM Persistent Area (PA) */
%let highest_fa_id       =; /* ID of the highest FA in the system. Example: If fa.3.1.7, provide 3.1.7 */
%let srchostname01       =; /* Host machine name */
%let srcirmport01        = 7980; /* Server port (defaulted to 7980) */
%let srcusername01       =; /* Username credentials */
%let srcpassword01       =; /* Password credentials (plain text or SAS-Encoded) */
/**************************************************************************************
    END OF REQUIRED PARAMETERS WHOSE VALUES MUST BE SET ON A PER SITE BASIS
 **************************************************************************************/

/**************************************************************************************
    OPTIONAL PARAMETERS WHOSE VALUES MAY BE SET ON A PER SITE BASIS
 **************************************************************************************/
%let debug        = TRUE; /* Change to FALSE if DEBUG info is not required */
%let logOptions   = ;      /* Logging options such as mprint symbolgen ...*/
%let pollInt      = 10;    /* Number of seconds between checks for status of instance. Defaulted to 10 */
%let maxWait      = 3600;  /* Max seconds to wait for an instance to complete. Defaulted to 3600 */
/**************************************************************************************
    END OF OPTIONAL PARAMETERS WHOSE VALUES MUST BE SET ON A PER SITE BASIS
 **************************************************************************************/

/* Determine script location. */
data _null_;
   set sashelp.vextfl;
   if (substr(fileref,1,3)='_LN' or substr(fileref,1,3)='#LN' or substr(fileref,1,3)='SYS') and
      index(upcase(xpath),'.SAS')>0 then do;
     call symputx("scr_path",substr(trim(xpath),1,(findc(trim(xpath),'/\','B'))-1));
      stop;
   end;
run;

%put Script is executing at &scr_path.;

/***********************************************************/
%let srcservertag    = 01;
%let myHostname      = &&srchostname&srcservertag.;
%let myUsername      = &&srcusername&srcservertag.;
%let myPassword      = &&srcpassword&srcservertag.;
%let myPort          = &&srcirmport&srcservertag.;
%let myRESTserver    = SASIRMServer; 
%let script_dir      = %scan(&scr_path.,-1,'/\');
%let script_loc      = %str(&scr_path.);
/***********************************************************/

/* Run setup macro to set SASAUTOS and LUA path */
%include "&scr_path./source/sas/ucmacros/irm_setup.sas";

%irm_setup( source_path = &scr_path.
          , fa_id       = &script_dir.
          );

/* Wrapper macro for re-creating unpublished flow instances */ 
%irm_rest_recreate_instances(
      host                = http://&myHostname.
   ,  port                = &myPort.
   ,  username            = &myUsername. 
   ,  password            = &myPassword.
   ,  server              = &myRESTserver.
   ,  debug               = &debug.
   ,  logOptions          = &logOptions.
   ,  final_fa_id         = &highest_fa_id.
   ,  pollInterval        = &pollInt.
   ,  maxWait             = &maxWait.
   ,  script_loc          = &script_loc.
   ,  pa_location         = &persistent_area_loc.
);

/* End of script to re-create instances. */
