--[[
/* Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA */

/*!
\file

\brief   Module provides JSON parsing functionality for interacting with the IRM REST API

\details 
   The following functions are available:
   - parseJSONFile(<filenamefileref>): Parse a JSON file and returns a Lua table
   - irmRestPlain(<filenamefileref>, output): Parse a JSON file and returns an output SAS table (Assumes a plain/flat JSON structure)

\ingroup commonAnalytics
 
\author  SAS Institute Inc.

\date    2015

*/

]]


local filesys = require 'sas.risk.utils.filesys'
local stringutils = require 'sas.risk.utils.stringutils'
local args = require 'sas.risk.utils.args'
local errors = require 'sas.risk.utils.errors'
local sas_msg = require 'sas.risk.utils.sas_msg'
local tableutils = require 'sas.risk.utils.tableutils'
local json = require 'plugin.dkjson'

local M = {}
-- declare function
local readFile
local parseJSONFile
local irmRestPlain

rowCount = function(ds)
   local dsid = sas.open(ds)
   local cnt = sas.nobs(dsid)
   sas.close(dsid)
   return cnt
end   
   
getJsonString = function(filename)
   -- Create a temporary table by reading the input file 1024 characters at the time
   local tmpDs = "__tmpJSON__"
   sas.set_quiet(true)
   sas.submit[[
      data @tmpDs@;
         infile @filename@ recfm = F lrecl = 1024 length = len;
         length str $1024.;
         input str $varying1024. len;
      run;  
   ]]

   -- Load data set
   local data = sas.load_ds(tmpDs)
   -- Load content of each row into a new table
   local rows = {}
   for i = 1, rowCount("__tmpJSON__") do
      rows[i] = data[i].str
   end
   
   -- Cleanup
   sas.submit[[
      proc datasets library = work nolist nodetails;
         delete @tmpDs@;
      quit;
   ]]
   sas.set_quiet(false)
   -- Concatenate the table to get a full JSON string
   local str = table.concat(rows)
   return str
end;


parseJSONFile = function(filename)
   --local jsonString = readFile(filename)
   local jsonString = getJsonString(filename)
   local jsonTable, pos, err
   jsonTable = nil
   if jsonString ~= "" then
      jsonTable, pos, err = json.decode (jsonString, 1, nil)
      if err then
         sas.set_quiet(true)
         sas.submit[[%put ERROR: @err@.;]]
         sas.set_quiet(false)
         --print ("ERROR:", err)
      end   
   else
      print("WARNING: Input filename is empty!")
   end   
   return jsonTable
end



processBoolean = function(tbl)
   for i, t in pairs(tbl) do
      if type(t) == "table" then
         tbl[i] = processBoolean(t)
      elseif type(t) == "boolean" then
         if t then
            tbl[i] = "true"
         else
            tbl[i] = "false"
         end
      end
   end
   return tbl
end;

irmRestPlain = function(filename, output)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      jsonTable = processBoolean(jsonTable)
      sas.write_ds(jsonTable, output)
   end
end;

M.irmRestPlain = irmRestPlain


irmRestEntities = function(filename, output)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      -- flatten the role sub-table to the same level as the entity list table
      for i, t in pairs(jsonTable) do
         for k, v in pairs(t.role) do
            t["role_"..k] = v
         end
         t.role = nil
      end;
      sas.write_ds(jsonTable, output)
   end
end;

M.irmRestEntities = irmRestEntities

irmRestInstances = function(filename, output)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      
      for i, t in pairs(jsonTable) do
         -- Remove the results and id fields
         t.results = nil
         t.id = nil
         -- Set the publish related info
         if t.publishInfo ~= nil then
            t.published = true
            t.publishedBy = t.publishInfo.createdBy
            t.publishInfo = nil
         else
            t.published = false
            t.publishedBy = ""
            t.publishComments = ""
            t.publishTimestamp = ""
         end
      end;
      jsonTable = processBoolean(jsonTable)
      sas.write_ds(jsonTable, output)
   end
end;
M.irmRestInstances = irmRestInstances

setJobStatus = function(status)
   res = "N/A"
   if status == 0 then
      res = "Not Run"
   elseif status == 2 then
      res = "Running"
   elseif status == 3 then
      res = "Error"
   elseif status == 4 then
      res = "Successful"
   elseif status == 6 then
      res = "Published"
   end
   return res
end

irmRestJobStatus = function(filename, output)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      result = {}

      -- Process jobflow status
      row = {}
      row.key = jsonTable.key
      row.type = "JOBFLOW"
      row.status_code = jsonTable.jobFlowStatus
      row.status = setJobStatus(jsonTable.jobFlowStatus)
      table.insert(result, row)
      
      -- Process subflow status
      for k, v in pairs(jsonTable.subFlowsStatus) do
         row = {}
         row.key = k
         row.type = "SUBFLOW"
         row.status_code = v
         row.status = setJobStatus(v)
         table.insert(result, row)
      end;
      
      -- Process task status
      for k, v in pairs(jsonTable.tasksStatus) do
         row = {}
         row.key = k
         row.type = "TASK"
         row.status_code = v
         row.status = setJobStatus(v)
         table.insert(result, row)
      end;
      sas.write_ds(result, output)
   end
end;
M.irmRestJobStatus = irmRestJobStatus


strScan = function (str, pos, pattern, plain)
   local outResults = {}
   local theStart = 1
   local theSplitStart, theSplitEnd = string.find(str, pattern, theStart, plain)
   local tokenCnt = 0
   while theSplitStart do
      table.insert(outResults, string.sub(str, theStart, theSplitStart-1))
      tokenCnt = tokenCnt + 1
      theStart = theSplitEnd + 1
      theSplitStart, theSplitEnd = string.find(str, pattern, theStart, plain)
   end
   table.insert(outResults, string.sub(str, theStart))
   tokenCnt = tokenCnt + 1
   
   if pos > 0 then
      return outResults[pos]
   elseif pos < 0 then
      return outResults[tokenCnt + pos + 1]
   else
      return outResults
   end
end


getObjectProperties = function(tbl, additionalProperties, subsetList, renameList)
   local res = {}
   for i, t in pairs(tbl) do
      -- Create new row
      local row = {}
      -- Attach the initial properties (assumption is that these properties have different name than those contained in the tbl object)
      for k, v in pairs(additionalProperties) do
         row[k] = v
      end
      -- Get all properties defined in the tbl object (current row)
      for k, v in pairs(t) do
         -- Keep only the properties that are specified in the subsetList
         if(subsetList[k] or subsetList[k] == nil) then
            if(renameList[k]) then
               row[renameList[k]] = v
            else
               row[k] = v
            end
         end
      end
      -- convert boolean to string
      row = processBoolean(row)
      -- Append this row to the result
      table.insert(res, row)
   end
   return res
end


irmRestTableMap = function(filename, output, webDavBasePath)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      local result = {}
      -- Rename the following properties in the output table
      local renameList = {processKey = "topInstanceKey", key = "dataKey", name = "objectName", baseName = "tableName"}
      -- Discard the following properties from the output table
      local subsetList = {id = false, sourceId = false, sourceTask = false}
      local columnList = {modified = "string"
                           , readOnly = "string"
                           , jobFlowDefinitionFile = "string"
                           , webDavNodePath = "string"
                           , dataKey = "number"
                           , nodeName = "string"
                           , collection = "string"
                           , url = "string"
                           , result = "string"
                           , webDavFullPath = "string"
                           , topInstanceKey = "number"
                           , categoryId = "string"
                           , baseDate = "string"
                           , webDavLibPath = "string"
                           , entityId = "string"
                           , owner = "string"
                           , fileName = "string"
                           , nodeType = "string"
                           , subflowKey = "number"
                           , libname = "string"
                           , version = "number"
                           , nodeKey = "number"
                           , tableName = "string"
                           , extension = "string"
                           , objectName = "string"
                           , partition = "string"
                          }
      -- Process subflow instances  
      for i, t in pairs(jsonTable.subFlowInstances) do
         local properties = {}
         if(not jsonTable.subFlow) then
            -- This is the root jobflow. Construct the base WebDAV path to the subflow folder
            properties.webDavNodePath = table.concat({jsonTable.baseDate
                                                  , jsonTable.configSetId
                                                  , jsonTable.entityId
                                                  , jsonTable.entityRoleId
                                                  , jsonTable.categoryId
                                                  , string.sub(jsonTable.jobFlowDefinitionFile,1,(string.len(jsonTable.jobFlowDefinitionFile)-5))
                                                  , jsonTable.name
                                                  , strScan(strScan(t.path, -1, "/", true), 1, ".", true)
                                                  }
                                                  , "/")
         else
            -- This is a subflow. The starting point is passed as a parameter by the call to the REST API
            properties.webDavNodePath = webDavBasePath .. "/" .. strScan(strScan(t.path, -1, "/", true), 1, ".", true)
         end
         properties.entityId = jsonTable.entityId
         properties.baseDate = jsonTable.baseDate
         properties.owner = jsonTable.createdBy
         properties.categoryId = jsonTable.categoryId
         properties.jobFlowDefinitionFile = jsonTable.jobFlowDefinitionFile

         properties.subflowKey = t.instanceKey
         properties.nodeKey = t.key
         properties.nodeName = t.name
         properties.nodeType = "SUBFLOW"
         -- Loop through all inputs
         local inputData = getObjectProperties(t.inputData, properties, subsetList, renameList)
         for idx, row in pairs(inputData) do
            row.webDavLibPath = table.concat({properties.webDavNodePath, row.libname}, "/")
            row.webDavFullPath = table.concat({properties.webDavNodePath, row.libname, row.fileName}, "/")
            table.insert(result, row)
         end
         -- Loop through all outputs
         local outputData = getObjectProperties(t.outputData, properties, subsetList, renameList)
         for idx, row in pairs(outputData) do
            row.webDavLibPath = table.concat({properties.webDavNodePath, row.libname}, "/")
            row.webDavFullPath = table.concat({properties.webDavNodePath, row.libname, row.fileName}, "/")
            table.insert(result, row)
         end
         -- Make sure this subflow object is included in the output table even if it does not have any input/output data object
         if (#inputData == 0 and #outputData == 0) then
            table.insert(result, properties)
         end
      end;
      
      -- Process task objects
      for i, t in pairs(jsonTable.tasks) do
         local properties = {}
         if(not jsonTable.subFlow) then
            -- This is the root jobflow. Construct the base WebDAV path to the task folder
            properties.webDavNodePath = table.concat({jsonTable.baseDate
                                                  , jsonTable.configSetId
                                                  , jsonTable.entityId
                                                  , jsonTable.entityRoleId
                                                  , jsonTable.categoryId
                                                  , strScan(jsonTable.jobFlowDefinitionFile, 1, ".", true)
                                                  , jsonTable.name
                                                  , t.name
                                                  }
                                                  , "/")
         else
            -- This is a subflow. The starting point is passed as a parameter by the call to the REST API
            properties.webDavNodePath = webDavBasePath .. "/" .. t.name
         end
         properties.entityId = jsonTable.entityId
         properties.baseDate = jsonTable.baseDate
         properties.owner = jsonTable.createdBy
         properties.categoryId = jsonTable.categoryId
         properties.jobFlowDefinitionFile = jsonTable.jobFlowDefinitionFile
         
         -- Make sure this property is defined in case of the task -> will result to missing column
         properties.subflowKey = -1/0
         properties.nodeKey = t.key
         properties.nodeName = t.name
         properties.nodeType = "TASK"
         -- Loop through all inputs
         local inputData = getObjectProperties(t.inputData, properties, subsetList, renameList)
         for idx, row in pairs(inputData) do
            row.webDavLibPath = table.concat({properties.webDavNodePath, row.libname}, "/")
            row.webDavFullPath = table.concat({properties.webDavNodePath, row.libname, row.fileName}, "/")
            table.insert(result, row)
         end
         -- Loop through all outputs
         local outputData = getObjectProperties(t.outputData, properties, subsetList, renameList)
         for idx, row in pairs(outputData) do
            row.webDavLibPath = table.concat({properties.webDavNodePath, row.libname}, "/")
            row.webDavFullPath = table.concat({properties.webDavNodePath, row.libname, row.fileName}, "/")
            table.insert(result, row)
         end
      end;

      -- Make sure that all columns are defined
      if(#result) then
         for k, v in pairs(columnList) do
            if(result[1][k] == nil) then
               if(v == "string") then
                  result[1][k] = ""
               else
                  result[1][k] = -1/0
               end
            end
         end
      end
      sas.write_ds(result, output)
   end
end
M.irmRestTableMap = irmRestTableMap

irmRestManageInstance = function(filename, output)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      -- Get rid of the results info
      jsonTable.results = nil
      jsonTable.id = nil
      if jsonTable.publishInfo ~= nil then
         jsonTable.published = true
         jsonTable.publishedBy = jsonTable.publishInfo.createdBy
         jsonTable.publishInfo = nil
      else
         jsonTable.published = false
         jsonTable.publishedBy = ""
         jsonTable.publishComments = ""
         jsonTable.publishTimestamp = ""
      end
      jsonTable = processBoolean(jsonTable)
      sas.write_ds({jsonTable}, output)
   end
end
M.irmRestManageInstance = irmRestManageInstance

irmRestMessages = function(filename, output)
   local jsonTable = parseJSONFile(filename)
   if jsonTable ~= nil then
      local result = {}
      -- Loop through all rows
      for i, row in pairs(jsonTable) do
         row.id = nil
         table.insert(result, row)
      end;
      sas.write_ds(result, output)
   end
end
M.irmRestMessages = irmRestMessages

return M
