/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Create a jobflow instance
   
   \param [in] instance_name Name of the jobflow instance to be created
   \param [in] description Jobflow description (optional)
   \param [in] entityId the entity id as returned by \link irm_rest_get_entities.sas \endlink (column: id)
   \param [in] entityRoleKey 1 (Solo) or 2(Group)
   \param [in] category The main category folder where the jobflows are stored
   \param [in] jobflowFile The bpmn file name (with or without .bpmn extension)
   \param [in] baseDate The base date (format: MMDDYYYY)
   \param [in] configSetId Configuration set id as returned by \link irm_rest_get_config_sets.sas \endlink (column: id)
   \param [in] federatedAreaID Id of the federated area as defined in the Metadata (Default: 0.3.1)
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] host Host url, including the protocol
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] port Server port (if blank, default port 80 will be used). Default: 7980
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] outvar Name of the output macro variable containing the instance key of the jobflow that has been created (Default: instanceKey)
   \param [in] outVarTicket Name of the output macro variable which will contain the Service Ticket (Default: ticket)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
      
   
   \details 
    This macro will create a jobflow instance and return the instance key (if successfull). The instance is automatically executed upon creation.
      
   <b>Example:</b>
   
   Create an jobflow istance for the Market SCR (factor based)
   \code
      %let instanceKey =;
      %irm_rest_create_jobflow_instance(instance_name = Market SCR Factor Based
                                        , description = Market SCR Factor Based approach
                                        , entityID = ENTITY_BE
                                        , entityRoleKey = 1
                                        , category = QRT - Solvency Capital Requirement 
                                        , jobflowFile = irmfw_S.26.01_solo_ann_conc_fb.bpmn
                                        , baseDate = 03312015
                                        , configSetId = EIOPA_FULL_SII_2.0.1
                                        , federatedAreaID = 0.3.1
                                        , host = <host>
                                        , port = <port>
                                        , username = <userid>
                                        , password = <pwd>
                                        , outvar = instanceKey
                                        );
      %put instanceKey = &instanceKey.;
   \endcode
         
         
   <b>Sample output:</b>

   \code
      instanceKey = 93888506
   \endcode
      
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/
%macro irm_rest_create_jobflow_instance(instance_name =
                                       , description =
                                       , entityId =
                                       , entityRoleKey =
                                       , category =
                                       , jobflowFile =  
                                       , baseDate =
                                       , configSetId =
                                       , federatedAreaID = %str(0.3.1)
                                       , LogonPort =
                                       , host =
                                       , server = SASIRMServer
                                       , port = 7980
                                       , username =
                                       , password =
                                       , outvar = instanceKey
                                       , outVarTicket = ticket
                                       , logOptions =
                                       );

   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;
   
   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   /* OutVarTicket cannot be missing. Set a default value */
   %if(%length(&outVarTicket.) = 0) %then
      %let outVarTicket = ticket;
   
   /* Declare the output variable as global if it does not exist */
   %if(not %symexist(&outVarTicket.)) %then
      %global &outVarTicket.;
                                                                              
   %if(%lowcase(%scan(&jobflowFile., -1, .)) ne bpmn) %then
      %let jobflowFile = &jobflowFile..bpmn;
   
   filename json_run temp;
   %put NOTE: Creating JSON payload for a new instance in &category - &instance_name.;
   proc json out = json_run pretty;
      write open o ; 
         write values "name" "&instance_name.";
         write values "description" "&description.";
         write values "entityRoleKey" "&entityRoleKey";
         write values "entityId" "&entityID.";
         write values "jobFlowDefinitionFile" "&jobflowFile.";
         write values "jobFlowDefinitionId" "&federatedAreaID.::%lowcase(&category.)::%lowcase(%sysfunc(prxchange(s/.bpmn$//i, -1, &jobflowFile.)))";
         write values "categoryId" "&category.";
         write values "federatedAreaId" "&federatedAreaID.";
         write values "baseDate" "&baseDate.";
         write values "configSetId" "&configSetID.";
      write close;
   run;

   %let url = %str(&host.:&port./&server./rest/jobflow/new);

   /* Reuse the ticket if it has been provided */
   %if(%length(&&&outVarTicket.) = 0) %then %do;   
      /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
      option nomlogic nosymbolgen;
      /* Get the Service Ticket */
      %irm_rest_get_ticket(url = &url.
                           , LogonPort = &LogonPort.
                           , server = &server.
                           , username = &username.
                           , password = &password.
                           , outVarTicket = &outVarTicket.
                           , logOptions = &oldLogOptions.
                           );
   %end;
   
   filename out_run temp;

   %put NOTE: Creation of a new instance via PROC HTTP;
   proc http
      in = json_run
      url = "%str(&url.)%nrstr(?&ticket=)&&&outVarTicket."
      out = out_run
      method = 'post'
      ct = 'application/json';
   run;
   
   data _null_;
      infile out_run;
      input ;
      call symput("&outvar.", _infile_);
   run;
   
   filename json_run clear;
   filename out_run clear;

%mend;
