/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Retrieve the list of Jobflow Instances from the server
   
   \param [in] host Host url, including the protocol
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] port Server port (if blank, default port 80 will be used). Default: 7980
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] outds Name of the output table (Default: instances_info)
   \param [in] outVarTicket Name of the output macro variable which will contain the Service Ticket (Default: ticket)
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
      
   \details 
   This macro sends a GET request to <b><i><host>:<port>/SASIRMServer/rest/jobflow/instancesinfo?entity=0&definition=0</i></b> and collects the results in the output table. \n
   \n
   See \link irm_rest_get_request.sas \endlink for details about how to send GET requests and parse the response.
      
   <b>Example:</b>
   
   1) Set up the environment (set SASAUTOS and required LUA libraries)
   \code
      %let source_path = <Path to the root folder of the Federated Content Area (root folder, excluding the Federated Content folder)>;
      %let fa_id = <Name of the Federated Area Content folder>;
      %include "&source_path./&fa_id./source/sas/ucmacros/irm_setup.sas";
      %irm_setup(source_path = &source_path.
                , fa_id = &fa_id.
                );
   \endcode

   2) Send a Http GET request and parse the JSON response into the output table WORK.instances_info
   \code
      %let ticket =;
      %irm_rest_get_instances(host = <host>
                              , port = <port>
                              , username = <userid>
                              , password = <pwd>
                              , outds = instances_info
                              , outVarTicket = ticket
                              );
   \endcode
         
   <b>Sample output:</b>

   | inputDataModified | published | configSetId          | key        | publishedBy | jobFlowDefinitionFile               | private | entityRoleKey | entityRoleId | federatedAreaId | categoryId                            | entityKey  | baseDate | lastRunTimestamp             | entityId  | publishTimestamp             | jobFlowDefinitionId                                                       | public | status | publishComments        | name                    | version | categoryKey | startup | visibility | createdBy | description |
   |-------------------|-----------|----------------------|------------|-------------|-------------------------------------|---------|---------------|--------------|-----------------|---------------------------------------|------------|----------|------------------------------|-----------|------------------------------|---------------------------------------------------------------------------|--------|--------|------------------------|-------------------------|---------|-------------|---------|------------|-----------|-------------|
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 1578684939 |             | irmfw_afb.bpmn                      | TRUE    | 1             | SOLO         | 0.3.1           | QRT - Submission Reports              | 1632778180 | 3312015  | 2015-10-20T17:29:03.464-0400 | ENTITY_BE |                              | 0.3.1::qrt - submission reports::irmfw_afb                                | FALSE  | 4      |                        | AFB                     | 1       | 1282754339  | FALSE   | PRIVATE    | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 1626468466 |             | irmfw_qrg.bpmn                      | TRUE    | 1             | GROUP        | 0.3.1           | QRT - Submission Reports              | 1467934750 | 3312015  | 2015-10-20T18:04:11.540-0400 | MAIN      |                              | 0.3.1::qrt - submission reports::irmfw_qrg                                | FALSE  | 4      |                        | QRG                     | 1       | 1282754339  | FALSE   | PRIVATE    | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 1028817144 |             | irmfw_ars.bpmn                      | TRUE    | 1             | SOLO         | 0.3.1           | QRT - Submission Reports              | 1632778180 | 3312015  | 2015-10-20T17:37:10.197-0400 | ENTITY_BE |                              | 0.3.1::qrt - submission reports::irmfw_ars                                | FALSE  | 4      |                        | ARS                     | 1       | 1282754339  | FALSE   | PRIVATE    | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 683261754  |             | irmfw_afb.bpmn                      | TRUE    | 1             | SOLO         | 0.3.1           | QRT - Submission Reports              | 1632778180 | 3312015  | 2015-10-22T02:56:36.169-0400 | ENTITY_BE |                              | 0.3.1::qrt - submission reports::irmfw_afb                                | FALSE  | 4      |                        | AFB flow 22OCT2015      | 1       | 1282754339  | FALSE   | PRIVATE    | sasdemo   |             |
   | FALSE             | TRUE      | EIOPA_FULL_SII_2.0.1 | 78709249   | sasdemo     | irmfw_S.26.01_solo_ann.bpmn         | FALSE   | 1             | SOLO         | 0.3.1           | QRT - Solvency Capital Requirement    | 1632778180 | 3312015  | 2015-10-21T11:02:35.722-0400 | ENTITY_BE | 2015-10-21T18:52:50.089-0400 | 0.3.1::qrt - solvency capital requirement::irmfw_s.26.01_solo_ann         | TRUE   | 6      | Aproved Market Results | S.26.01.01              | 1       | 2005090872  | FALSE   | PUBLIC     | sasdemo   |             |
   | FALSE             | TRUE      | EIOPA_FULL_SII_2.0.1 | 81464256   | sasdemo     | irmfw_S.06.01_solo_ann.bpmn         | FALSE   | 1             | SOLO         | 0.3.1           | QRT - Assets and Investments          | 1632778180 | 3312015  | 2015-10-20T17:13:22.357-0400 | ENTITY_BE | 2015-10-22T17:20:46.014-0400 | 0.3.1::qrt - assets and investments::irmfw_s.06.01_solo_ann               | TRUE   | 6      | Final list of assets   | S.06.01.01 - test       | 1       | 1569236270  | FALSE   | PUBLIC     | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 804639510  |             | irmfw_S.28.01.01.bpmn               | TRUE    | 1             | SOLO         | 0.3.1           | QRT - Minimum Capital Requirement     | 1632778180 | 3312015  | 2015-10-20T17:20:16.834-0400 | ENTITY_BE |                              | 0.3.1::qrt - minimum capital requirement::irmfw_s.28.01.01                | FALSE  | 4      |                        | MCR Reports             | 1       | 1959708959  | FALSE   | PRIVATE    | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 767689337  |             | irmfw_qs_qrs_qrt.bpmn               | FALSE   | 1             | SOLO         | 0.3.1           | QRT - Submission Reports (Quickstart) | 1632778180 | 3312015  | 2015-10-21T07:43:14.312-0400 | ENTITY_BE |                              | 0.3.1::qrt - submission reports (quickstart)::irmfw_qs_qrs_qrt            | TRUE   | 4      |                        | QuickStart QRS          | 1       | 1730609353  | FALSE   | PUBLIC     | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 1904186996 |             | irmfw_arg.bpmn                      | TRUE    | 2             | GROUP        | 0.3.1           | QRT - Submission Reports              | 1467934750 | 3312015  | 2015-10-21T11:30:07.942-0400 | MAIN      |                              | 0.3.1::qrt - submission reports::irmfw_arg                                | FALSE  | 4      |                        | ARG                     | 1       | 1282754339  | FALSE   | PRIVATE    | sasdemo   |             |
   | FALSE             | FALSE     | EIOPA_FULL_SII_2.0.1 | 1596921364 |             | irmfw_S.26.01_solo_ann_conc_fb.bpmn | TRUE    | 1             | SOLO         | 0.3.1           | QRT - Solvency Capital Requirement    | 1632778180 | 3312015  | 2015-10-21T08:05:39.737-0400 | ENTITY_BE |                              | 0.3.1::qrt - solvency capital requirement::irmfw_s.26.01_solo_ann_conc_fb | FALSE  | 4      |                        | Market SCR Factor Based | 1       | 2005090872  | FALSE   | PRIVATE    | sasdemo   |             |   
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/
%macro irm_rest_get_instances(host =
                             , LogonPort =
                             , server = SASIRMServer
                             , port = 7980
                             , username =
                             , password =
                             , outds = instances_info
                             , outVarTicket = ticket
                             , debug = false
                             , logOptions =
                             );
                             
   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   %if(%length(&port.) = 0) %then
      %let port = 80;

   /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
   option nomlogic nosymbolgen;
   /* Send the REST request */
   %irm_rest_get_request(url = &host:&port./&server./rest/jobflow/instancesinfo?entity=0%str(&)definition=0
                         , LogonPort = &LogonPort.
                         , server = &server.
                         , username = &username.
                         , password = &password.
                         , parser = irmRestInstances
                         , outds = &outds.
                         , outVarTicket = &outVarTicket.
                         , debug = &debug.
                         , logOptions = &oldLogOptions.
                         );
   
%mend;
