/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Send Http GET request to the IRM REST API and parse the result
   
   \param [in] url Full url to the REST resource 
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] parser Name of the LUA parser function responsible for converting the JSON response into the output SAS table (Default: irmRestPlain -> plain one-level JSON structure). See \link json.lua \endlink for details).
   \param [in] outds Name of the output table
   \param [in] outVarTicket Name of the output macro variable which will contain the Service Ticket (Default: ticket)
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
   \param [in] arg1 Additional parameter passed to the LUA parser function
   \param [in] arg2 Additional parameter passed to the LUA parser function
   \param [in] arg3 Additional parameter passed to the LUA parser function
   \param [in] arg4 Additional parameter passed to the LUA parser function
   \param [in] arg5 Additional parameter passed to the LUA parser function
   
   
   
   \details 
   This macro performs the following operations:
      1. Get a Service Ticket for the requested URL (see \link irm_rest_get_ticket.sas \endlink for details)
         1.a the ticket is created only if the macro variable referenced by outVarTicket is blank (i.e. -> <i> if &&&outVarTicket = %str() </i>)
      2. Send Http GET request for the specified URL
      3. Parse the JSON response using the specified parser function (only if the parser is provided)
      4. Return the parsed response into a SAS table. 
      

   <b>Example:</b>
   
   1) Set up the environment (set SASAUTOS and required LUA libraries)
   \code
      %let source_path = <Path to the root folder of the Federated Content Area (root folder, excluding the Federated Content folder)>;
      %let fa_id = <Name of the Federated Area Content folder>;
      %include "&source_path./&fa_id./source/sas/ucmacros/irm_setup.sas";
      %irm_setup(source_path = &source_path.
                , fa_id = &fa_id.
                );
   \endcode
   
   2) Send a Http GET request and parse the JSON response into the output table WORK.result using the default parser irmRestPlain
   \code
      %let ticket =;
      %irm_rest_get_request(url = <host>:<port>/SASIRMServer/<resource>
                            , server = SASIRMServer
                            , username = <userid>
                            , password = <pwd>
                            , parser = irmRestPlain
                            , outds = result
                            , outVarTicket = ticket
                            );   
   \endcode
   
   The following utility macros make use of this macro: 
    - List of Base dates: \link irm_rest_get_basedates.sas \endlink
    - List of Jobflow categories: \link irm_rest_get_categories.sas \endlink
    - List of Configuration Sets: \link irm_rest_get_config_sets.sas \endlink
    - List of Entities: \link irm_rest_get_entities.sas \endlink
    - List of Jobflow definitions: \link irm_rest_get_flow_def.sas \endlink
    - List of Jobflow instances: \link irm_rest_get_instances.sas \endlink
    - Status of a Jobflow instance: \link irm_rest_get_job_status.sas \endlink
    - List of all data objects of a Jobflow instance: \link irm_rest_get_table_map.sas \endlink
    
      
   \note The structure of the output table depends on the JSON content and the parser function
         
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/

%macro irm_rest_get_request(url =
                            , LogonPort =
                            , server = SASIRMServer
                            , username =
                            , password =
                            , parser = irmRestPlain
                            , outds = 
                            , outVarTicket = ticket
                            , debug = false
                            , logOptions =
                            , arg1 =
                            , arg2 =
                            , arg3 =
                            , arg4 =
                            , arg5 =
                            );

   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   /* Delete output table if it exists */
   %if(%rsk_dsexist(&outds.)) %then %do;
      proc sql;
         drop table &outds.;
      quit;
   %end;
   
   /* OutVarTicket cannot be missing. Set a default value */
   %if(%length(&outVarTicket.) = 0) %then
      %let outVarTicket = ticket;
   
   /* Declare the output variable as global if it does not exist */
   %if(not %symexist(&outVarTicket.)) %then
      %global &outVarTicket.;
   
   /* Reuse the ticket if it has been provided */
   %if(%length(&&&outVarTicket.) = 0) %then %do;
      /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
      option nomlogic nosymbolgen;
      /* Get the Service Ticket */
      %irm_rest_get_ticket(url = &url.
                           , LogonPort = &LogonPort.
                           , server = &server.
                           , username = &username.
                           , password = &password.
                           , outVarTicket = &outVarTicket.
                           , debug = &debug.
                           , logOptions = &oldLogOptions.
                           );
   %end;

   %local question_mark;
   %if(%index(&url.,?) = 0) %then
      %let question_mark = ?;
      
   filename out temp;
   %put NOTE: Getting the instance key via PROC HTTP;
   %if (%upcase(&debug.) eq TRUE) %then %do;
      %put ------------------------------------------------;
      %put REST Request: %str(&url.)&question_mark.%nrstr(&ticket=)&&&outVarTicket.;
      %put ------------------------------------------------;
   %end;
   
   proc http
      url = "%str(&url.)&question_mark.%nrstr(&ticket=)&&&outVarTicket."
      out = out
      method = 'GET'
      ct = 'application/json';
   run;

   /* Parse the request if a JSON parser is provided */
   %if (%length(&parser.)) %then %do;
      /* Declare the PRODUCT macro variable, required in sas.risk.rmx.rsk_init @ line # 24 */
      %local PRODUCT;
      %let PRODUCT = IRM; 
      /* Parse the JSON file and write the result in the output table */
      %rsk_call_riskexec(entrypoint_module         =   sas.risk.irm.json
                         , entrypoint_function     =   &parser.
                         , arg1                    =   out
                         , arg2                    =   &outds.
                         , arg3                    =   &arg1.
                         , arg4                    =   &arg2.
                         , arg5                    =   &arg3.
                         , arg6                    =   &arg4.
                         , arg7                    =   &arg5.
                         );
   %end;
   
   %if(%upcase(&debug.) ne TRUE) %then %do;
      filename out clear;
   %end;
%mend;
