/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Retrieve the list of all data objects of a Jobflow instance from the server
   
   \param [in] instanceKey Unique identifier of the instance
   \param [in] recursive True/False If True, the jobflow instance is recursively traversed to collect info about each subflow (Default: true).
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] host Host url, including the protocol
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] port Server port (if blank, default port 80 will be used). Default: 7980
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] outds Name of the output table (Default: jobflow_table_map)
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
      
   
   \details 
   This macro sends a GET request to <b><i><host>:<port>/SASIRMServer/rest/jobflow/instance/&InstanceKey.</i></b> and collects the results in the output table. \n
   If <b><i>recursive = true</i></b>, an additional request is sent for each subflow that is found.\n
   \n
   See \link irm_rest_get_request.sas \endlink for details about how to send GET requests and parse the response.
      
   <b>Example:</b>
   
   1) Set up the environment (set SASAUTOS and required LUA libraries)
   \code
      %let source_path = <Path to the root folder of the Federated Content Area (root folder, excluding the Federated Content folder)>;
      %let fa_id = <Name of the Federated Area Content folder>;
      %include "&source_path./&fa_id./source/sas/ucmacros/irm_setup.sas";
      %irm_setup(source_path = &source_path.
                , fa_id = &fa_id.
                );
   \endcode

   2) Send a Http GET request and parse the JSON response into the output table WORK.jobflow_table_map
   \code
      %irm_rest_get_table_map(instanceKey = <instance key>
                              , recursive = false
                              , host = <host>
                              , port = <port>
                              , username = <userid>
                              , password = <pwd>
                              , outds = jobflow_table_map
                              );
   \endcode
         
   <b>Sample output:</b>

   | modified | readOnly | jobFlowDefinitionFile               | webDavNodePath                                                                                                         | dataKey    | nodeName              | collection | url                                                             | result | webDavFullPath                                                                                                                                           | topInstanceKey | categoryId                         | baseDate | webDavLibPath                                                                                                                 | entityId  | owner   | fileName                   | nodeType | subflowKey | libname | version | nodeKey    | tableName         | extension | objectName                        | partition |
   |----------|----------|-------------------------------------|------------------------------------------------------------------------------------------------------------------------|------------|-----------------------|------------|-----------------------------------------------------------------|--------|----------------------------------------------------------------------------------------------------------------------------------------------------------|----------------|------------------------------------|----------|-------------------------------------------------------------------------------------------------------------------------------|-----------|---------|----------------------------|----------|------------|---------|---------|------------|-------------------|-----------|-----------------------------------|-----------|
   | FALSE    | FALSE    | irmfw_S.26.01_solo_ann_conc_fb.bpmn | 03312015/EIOPA_FULL_SII_2.0.1/ENTITY_BE/SOLO/QRT - Solvency Capital Requirement/irmfw_S/Market SCR/irmfw_scr_m_conc_fb | 1105330724 | Market SCR            | FALSE      | irmcontent/pa/data/1596921364/mk_scr/market_cap_charge.sas7bdat | FALSE  | 03312015/EIOPA_FULL_SII_2.0.1/ENTITY_BE/SOLO/QRT - Solvency Capital Requirement/irmfw_S/Market SCR/irmfw_scr_m_conc_fb/mk_scr/market_cap_charge.sas7bdat | 1596921364     | QRT - Solvency Capital Requirement | 3312015  | 03312015/EIOPA_FULL_SII_2.0.1/ENTITY_BE/SOLO/QRT - Solvency Capital Requirement/irmfw_S/Market SCR/irmfw_scr_m_conc_fb/mk_scr | ENTITY_BE | sasdemo | market_cap_charge.sas7bdat | SUBFLOW  | 1449440686 | mk_scr  | 1       | 1113517225 | market_cap_charge | sas7bdat  | MK_SCR.MARKET_CAP_CHARGE.SAS7BDAT | FALSE     |
   | FALSE    | FALSE    | irmfw_S.26.01_solo_ann_conc_fb.bpmn | 03312015/EIOPA_FULL_SII_2.0.1/ENTITY_BE/SOLO/QRT - Solvency Capital Requirement/irmfw_S/Market SCR/S                   | 2032628613 | S.26.01 QRTgeneration | FALSE      | irmcontent/pa/data/1596921364/output/S.26.01.01.xlsx            | TRUE   | 03312015/EIOPA_FULL_SII_2.0.1/ENTITY_BE/SOLO/QRT - Solvency Capital Requirement/irmfw_S/Market SCR/S/output/S.26.01.01.xlsx                              | 1596921364     | QRT - Solvency Capital Requirement | 3312015  | 03312015/EIOPA_FULL_SII_2.0.1/ENTITY_BE/SOLO/QRT - Solvency Capital Requirement/irmfw_S/Market SCR/S/output                   | ENTITY_BE | sasdemo | S.26.01.01.xlsx            | SUBFLOW  | 621631133  | output  | 1       | 484874770  | S.26.01.01        | xlsx      | OUTPUT.S.26.01.01.xlsx            | FALSE     |
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/
%macro irm_rest_get_table_map(instanceKey =
                              , recursive = true
                              , host =
                              , LogonPort =
                              , server = SASIRMServer
                              , port = 7980
                              , username =
                              , password =
                              , outds = jobflow_table_map
                              , debug = false
                              , logOptions =
                              );

   %local 
      ticket
      currentKey
      webDavPath
      done
   ;
   
   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   %if(%length(&port.) = 0) %then
      %let port = 80;

   %if(%rsk_dsexist(&outds.)) %then %do;
      proc sql nowarn;
         drop table &outds.;
      quit;
   %end;
   
   /* Insert the top instance key into the queue */
   data _tmp_rest_queue;
      length 
         instanceKey 8.
         nodeName $500.
         webDavNodePath $500.
      ;
      instanceKey = &instanceKey.;
   run;
   
   %let done = N;
   %do %while(&done. eq N);
      /* Check if there are any flows/subflows to process */
      %if(%rsk_attrn(_tmp_rest_queue, nobs)) %then %do;
         %let currentKey =;
         %let webDavPath =;
         data _tmp_rest_queue;
            set _tmp_rest_queue;
            /* Get the next instance from the queue */
            if _N_ = 1 then do;
               call symputx("currentKey", instanceKey, "L");
               call symputx("webDavNodePath", webDavNodePath, "L");
            end;
            if _N_ > 1;
         run;

         proc datasets library = work nolist nodetails nowarn;
            delete _tmp_table_map;
         quit;

         /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
         option nomlogic nosymbolgen;
         /* Get the jobflow info (the ticket is reused on subsequent calls) */
         %irm_rest_get_request(url = &host:&port./&server./rest/jobflow/instance/&currentKey.
                               , LogonPort = &LogonPort.
                               , server = &server.
                               , username = &username.
                               , password = &password.
                               , parser = irmRestTableMap
                               , outds = _tmp_table_map
                               , outVarTicket = ticket
                               , debug = &debug.
                               , logOptions = &oldLogOptions.
                               , arg1 = &webDavNodePath.
                               );

         /* Check if the call to the REST API produced any output */
         %if(%rsk_dsexist(_tmp_table_map)) %then %do;
            proc append data = _tmp_table_map 
                        base = &outds. force;
            run;
            
            proc sort data = _tmp_table_map(keep = subflowKey nodeName webDavNodePath
                                            where = (subflowKey is not missing)
                                            )
                      out = _tmp_subflow_list (rename = (subflowKey = instanceKey)) nodupkey;
               by subflowKey;
            run;
                        
            /* Get the list of subflows and append them to the queue */
            proc append data = _tmp_subflow_list
                        base = _tmp_rest_queue force;
            run;
            
            %if(%rsk_attrn(_tmp_subflow_list, nobs)) %then %do;
               %put --------------------------------------------------------------------------------------;
               %put Found %rsk_attrn(_tmp_subflow_list, nobs) subflows. Adding them to the queue (total in queue: %rsk_attrn(_tmp_rest_queue, nobs));
               %put --------------------------------------------------------------------------------------;
            %end;

         %end;
         
         /* Check if we we need to recursively collect info about subflows */
         %if(%upcase(&recursive.) ne TRUE) %then
            %let done = Y;
         
      %end; /* %if(%rsk_attrn(_tmp_rest_queue, nobs)) */
      %else
         %let done = Y;
      
   %end; /* %do %while(&done. eq N); */
   
   %if(%upcase(&debug.) ne TRUE) %then %do;
      proc datasets library = work nolist nodetails nowarn;
         delete 
            _tmp_subflow_list
            _tmp_rest_queue
            _tmp_table_map
         ;
      quit;
   %end;
   
%mend;
