/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Authenticate and retrieve IRM REST API service ticket using Central Authentication Service (CAS)
   
   \param [in] url Full url to the REST resource 
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] outVarTicket Name of the output macro variable which will contain the Service Ticket (Default: service_ticket)
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
   
   
   \details 
   The Service Ticket is an authorizing ticket which allows a user to login into the application having \<Service URL\> as a target destination. 
   This is a ticket obtained by explicitly providing user credentials in some form (REST API/Java Service Registry -AuthenticationClient & AuthenticationClientHolder)
   \n
   
   This macro performs the following operations:
      1. Initially use the SASLogon webapp to acquire a Ticket Granting Ticket (TGT)
      2. Use the TGT to request a Service Ticket (ST) in order to access a specific SAS REST service API.
   \n \n
   
   Once a Service Ticket for a given <i>\<URL\></i> has been acquired, the target resource can be obtained by sending a http request to <i>\<URL\><b>?&amp;ticket=\<Service Ticket\></b></i>.

   <b>Example:</b>
   
   1) Set up the environment (set SASAUTOS and required LUA libraries)
   \code
      %let source_path = <Path to the root folder of the Federated Content Area (root folder, excluding the Federated Content folder)>;
      %let fa_id = <Name of the Federated Area Content folder>;
      %include "&source_path./&fa_id./source/sas/ucmacros/irm_setup.sas";
      %irm_setup(source_path = &source_path.
                , fa_id = &fa_id.
                );
   \endcode
   
   2) Get the service ticket for a specific resource
   \code
      %let service_ticket =;
      %let url = <protocol>://<host>:<port>/SASIRMServer/rest/basedates;
      %irm_rest_get_ticket(url = &url.
                           , username = <userid>
                           , password = <pwd>
                           , outVarTicket = service_ticket
                           );
      %put service_ticket = &service_ticket.;
   \endcode
   
   3) then use the ticket to request the resource and print the result to the log
   \code
      filename out temp;
      proc http
         url = "&url.?%str(&)ticket=&service_ticket."
         out = out
         method = 'GET'
         ct = 'application/json';
      run;
      
      data _null_;
         infile out;
         input;
         put _infile_;
      run;
      
      filename out clear;
   \endcode

   <b>Sample output:</b>
   \code
      [
        {
          "current": true,
          "id": "03312015",
          "version": 1,
          "name": "03312015",
          "key": 1
        },
        {
          "current": false,
          "id": "03312014",
          "version": 1,
          "name": "03312014",
          "key": 2
        }
      ]   
   \endcode
   
   \note The client needs to acquire a ST for each different service URL (not simply for each service context or resource path). \n
          Normally, CAS service tickets are <b>one-time-use only</b>, but for the SAS implementation, a ST can be reused when accessing the same URL.
         
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/

%macro irm_rest_get_ticket(url =
                           , LogonPort =
                           , server = SASIRMServer
                           , username = %nrstr()
                           , password = %nrstr()
                           , outVarTicket = service_ticket
                           , debug = false
                           , logOptions =
                           );
                           
   %local 
      oldnotes
      oldsgen
      oldmprint
      oldmlogic
      i
   ;
   
   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;
   
   %let oldnotes = %sysfunc(getoption(notes));
   %let oldsgen = %sysfunc(getoption(symbolgen));
   %let oldmprint = %sysfunc(getoption(mprint));
   %let oldmlogic = %sysfunc(getoption(mlogic));

   /* Use regex to extract the protocol and host from the url parameter */
   %if(%length(&LogonPort.) > 0) %then
      %let url2 = %sysfunc(prxchange(s/(https?:\/\/[^:]+)(:(\d+))?(\/&server.)(.*)/$1:&LogonPort./i, -1, %superq(url)));  
   %else
      /* &LogonPort is missing: assume that the /SASLogon web application is on the same port of /SASIRMServer */
      %let url2 = %sysfunc(prxchange(s/(https?:\/\/[^:]+)(:(\d+))?(\/&server.)(.*)/$1$2/i, -1, %superq(url)));
   
   %if (%upcase(&DEBUG.) EQ TRUE) %then %do; 
      %put Server URL: &url2.;
   %end;

   /* Declare the output variable as global if it does not exist */
   %if not %symexist(&outVarTicket.) %then
      %global &outVarTicket.;
   ;
   
   /* get the TGT */
   filename hin  temp;
   filename hout temp;
   filename in   temp;
   filename out  temp;

   data _null_;
      file hin;
      put 'accept:text/plain';
   run;

   /* Suppressing temporarily the printing options to the log */
   option nonotes nosgen nomprint nomlogic;   


   proc http 
         url       = %str("&url2./SASLogon/v1/tickets")
         headerin  = hin 
         headerout = hout 
         in        = "username=%bquote(&username.)%nrstr(&)password=%bquote(&password.)"
         method    = 'post' 
         ct        = 'text/plain'
         ;
   run;
   
   /* Reactivate the log options */
   option &oldnotes. &oldsgen. &oldmprint. &oldmlogic.;

   data _null_;
      infile hout;
      input;

      retain rx;
      if _N_ = 1 then
         rx = prxparse("s/^Location:\s*//i");
      
      if prxmatch(rx, _infile_) then do;
         /* Strip the initial portion "Location:" from the string */
         tgt = prxchange(rx, -1, _infile_);
         call symput("&outVarTicket", tgt); 
         stop;
      end;
   run;

   %if (%upcase(&DEBUG.) EQ TRUE) %then %do; 
      %put &outVarTicket. = &&&outVarTicket.; 
   %end;

   %if %index(&&&outVarTicket., %str(TGT-)) EQ 0 %then %do;
      %put ERROR: Failed to get ticket grant ticket for IRM REST API.;
      %return;
   %end;
   
   /* get the ST */
   data _null_;
      file in recfm = f lrecl = 1;
      put "service=&url";
   run;

   proc http 
         url       = %str("%trim(&&&outVarTicket..)")
         headerin  = hin 
         headerout = hout
         in        = in
         out       = out
         method    = 'post' 
         ct        = 'text/plain'
         ;
   run;

   /* get the ST */
   data _null_;
      infile out;
      input;
      call symput("&outVarTicket.", _infile_); 
   run;

   %if (%upcase(&DEBUG.) EQ TRUE) %then %do; 
      %put -----------------------------------------------;
      %put &outVarTicket = &&&outVarTicket..; 
      %put -----------------------------------------------;
   %end;

   %if %index(&&&outVarTicket.., %str(ST-)) NE 1 %then %do;
      %put ERROR: Failed to get service ticket for IRM REST API.;
      %return;
   %end;
   
   filename in clear;  
   filename out clear;
   filename hin clear;
   filename hout clear;
   
%mend;
