/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Perform an action (Share/Unshare/Publish/Delete) on a Jobflow instance
   
   \param [in] instanceKey Unique identifier of the instance
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] host Host url, including the protocol
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] port Server port (if blank, default port 80 will be used). Default: 7980
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] outds Name of the output table (Default: status)
   \param [in] action One of: Share/Unshare/Publish/Delete.
   \param [in] comment A comment to be used in case of publishing
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
   
   
   
   \details 
   Depending on the selected action, on of the following requests is sent to the server:
      - Share: <i><host>:<port>/SASIRMServer/rest/jobflow/instances/&instanceKey./share (Http verb: PUT)
      - Unshare: <i><host>:<port>/SASIRMServer/rest/jobflow/instances/&instanceKey./unshare (Http verb: PUT)
      - Publish: <i><host>:<port>/SASIRMServer/rest/jobflow/instances/&instanceKey./publish (Http verb: POST)
      - Delete: <i><host>:<port>/SASIRMServer/rest/jobflow/instance/&instanceKey</i> (Http verb: DELETE)
      
   <b>Example:</b>
   
   1) Set up the environment (set SASAUTOS and required LUA libraries)
   \code
      %let source_path = <Path to the root folder of the Federated Content Area (root folder, excluding the Federated Content folder)>;
      %let fa_id = <Name of the Federated Area Content folder>;
      %include "&source_path./&fa_id./source/sas/ucmacros/irm_setup.sas";
      %irm_setup(source_path = &source_path.
                , fa_id = &fa_id.
                );
   \endcode

   2) Publish an instance and parse the server JSON response into the output table WORK.status
   \code
      %irm_rest_manage_instance(instanceKey = <instance key>
                                , host = <host>
                                , port = <port>
                                , username = <userid>
                                , password = <pwd>
                                , outds = status
                                , action = publish
                                , comment = "Publish results"
                                );
   \endcode
         
         
   <b>Sample output:</b>

   | jobFlowDefinitionId                                                | inputDataModified | published | configSetId          | version | key        | publishedBy | private | entityRoleKey | entityRoleId | federatedAreaId | publishTimestamp             | entityKey  | baseDate | lastRunTimestamp             | entityId  | categoryId                         | public | description            | startup | publishComments | name       | status | categoryKey | jobFlowDefinitionFile        | visibility | createdBy |
   |--------------------------------------------------------------------|-------------------|-----------|----------------------|---------|------------|-------------|---------|---------------|--------------|-----------------|------------------------------|------------|----------|------------------------------|-----------|------------------------------------|--------|------------------------|---------|-----------------|------------|--------|-------------|------------------------------|------------|-----------|
   | 0.3.1::qrt - solvency capital requirement::irmfw_sr.26.01_both_ann | FALSE             | TRUE      | EIOPA_FULL_SII_2.0.1 | 1       | 1360731541 | sasdemo     | FALSE   | 1             | SOLO         | 0.3.1           | 2015-10-23T12:19:29.143-0400 | 1632778180 | 3312015  | 2015-10-23T10:54:59.616-0400 | ENTITY_BE | QRT - Solvency Capital Requirement | TRUE   | Market SCR Calculation | FALSE   | Publish results | Market SCR | 6      | 2005090872  | irmfw_SR.26.01_both_ann.bpmn | PUBLIC     | sasdemo   |
   
   \note Output table <b>is not produced</b> in case of DELETE action.
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/
%macro irm_rest_manage_instance(instanceKey =
                               , host =
                               , LogonPort =
                               , server = SASIRMServer
                               , port = 7980
                               , username =
                               , password =
                               , outds = status
                               , action = /* share/unshare/publish/delete */
                               , comment =
                               , debug = false
                               , logOptions =
                               ) / minoperator;

   %local 
      request_method
   ;
   
   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   %if(%length(&port.) = 0) %then
      %let port = 80;

   %if(not (%upcase(&action.) in SHARE UNSHARE PUBLISH DELETE)) %then %do;
      %put ERROR: Found ACTION = &ACTION.. Expected values: SHARE/UNSHARE/PUBLISH/DELETE;
      %goto EXIT;
   %end;
      
   %if(%upcase(&action.) EQ DELETE) %then %do;
      %let request_method = DELETE;
      %let url = %str(&host.:&port./&server./rest/jobflow/instance/&instanceKey.);
   %end;
   %else %do;
      %if (%upcase(&action.) EQ PUBLISH) %then
         %let request_method = POST;
      %else
         %let request_method = PUT;
      %let url = %str(&host.:&port./&server./rest/jobflow/instances/&instanceKey./%lowcase(&action.));
   %end;

   /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
   option nomlogic nosymbolgen;
   
   %local ticket;
   %irm_rest_get_ticket(url = &url.
                        , LogonPort = &LogonPort.
                        , server = &server.
                        , username = &username.
                        , password = &password.
                        , outVarTicket = ticket
                        , debug = &debug.
                        , logOptions = &oldLogOptions.
                        );

   filename in temp;
   filename out temp;

   data _null_;
      file in;
      put "{""comment"":""&comment.""}";
   run;
   
   proc http
      url = "%str(&url.)%nrstr(?&ticket=)&ticket"
      in = in
      %if(&request_method. eq PUT) %then %do;
         /* Use java_http method else it would not produce any output file when using the PUT method */
         java_http
      %end;
      out = out
      method = "&request_method."
      ct = "application/json";
   run;

   /* Proc HTTP does not support output in case of DELETE method */
   %if(&request_method. ne DELETE) %then %do;
      /* Declare the PRODUCT macro variable, required in sas.risk.rmx.rsk_init @ line # 24 */
      %local PRODUCT;
      %let PRODUCT = IRM; 
      /* Parse the JSON file and write the result in the output table */
      %rsk_call_riskexec(entrypoint_module         =   sas.risk.irm.json
                         , entrypoint_function     =   irmRestManageInstance
                         , arg1                    =   out
                         , arg2                    =   &outds.
                         );

   %end;
   %else %do;
      %put Delete request for instance &instanceKey has been sent.;
      /* Better way would be to send another get request to check if the instance still exists or it has been deleted */
   %end;

   
   %if(%upcase(&debug.) ne TRUE) %then %do;
      filename in clear;
      filename out clear;
   %end;

   %EXIT:
                         
%mend;
