/*
 Copyright (C) 2017 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   The macro is invoked from the script recreate_flow_instances.sas to delete unpublished instances (status="COMPLETED_SUCCESS") and re-create the instances. 
      
   \param [in] host           Host machine name. Example: irmserver.abc.xyz.com
   \param [in] port           Server port. Default: 7980 
   \param [in] username       Username credentials of user that created instances. Instances will be deleted and re-created for the user. Example: irmuser1
   \param [in] password       Password credentials for the user above. Can be plain text or SAS-Encoded (will be masked during execution).
   \param [in] server         Name of the Web Application Server that provides the REST service. Default: SASIRMServer
   \param [in] debug          Amount of execution information in the log. Defaulted to TRUE. Change to FALSE if DEBUG info is not required in the log.
   \param [in] logOptions     Amount of macro execution information in the log. Defaulted to blank. Could be changed to mprint mlogic symbolgen and other SAS supported options.
   \param [in] final_fa_id    ID of the highest Federated Area (FA) in the system. The FA contains all information for instance re-creation. New instances will be associated with the FA ID.
   \param [in] pollInterval   Number of seconds between checks for status of instance creation. Defaulted to 10 seconds.
   \param [in] maxWait        Maximum seconds to wait for an instance creation to complete. Default to 3600 seconds.        
   \param [in] script_loc     Location where the script is executing from. Example: /local/install/IRM/SAS94/recreateInstanceJob        
   \param [in] pa_location    Complete path to the IRM Persistent Area (PA). Example: /local/install/IRM/SAS94/PersistentArea        
   
   \details 
   The algorithm of the macro is as follows:
   1. Macro execution starts with the printing of the line "ENTERING IRM_REST_RECREATE_INSTANCES" to the execution log.
   2. For the specified user, retrieve the list of existing instances. This is done my making a REST call with the macro \link irm_rest_get_instances.sas
   3. A html file by the naming convention "existing_instances_<user>.html" is created at the script execution location. The file contains a listing of retrieved instances.
   4. Only instances with status of "COMPLETED_SUCCESS" are selected and separated out into three SAS datasets based on whether they are Published, contain Modified inputs or, none of the two.
   5. Instances with modified inputs are processed by invoking the macro \link irm_rest_run_mod_instance.sas. The macro collects uploaded inputs to the instance, copies them to a temporary location, deletes the 
      current instance, re-creates the instance and uploads the modified inputs.
   6. Published instances are retained "as-is". All other instances are processed by invoking the macros \link irm_manage_instances.sas (deletes the current instance) and \link irm_rest_create_jobflow_instance.sas (re-creates 
      the instance with information retrieved for the instance in Step 1. The macro \link irm_rest_wait_for_job.sas  polls the server to check if the instance re-creation is complete before processing the next instance.
      New instance keys are written to the execution log preceded by the string "NewInstanceKey =" 
   7. After all eligible instances have been re-created, the macro \link irm_rest_get_instances.sas is invoked again to retrieve all instances for the specified user at the end of instance re-creation. 
      A html file by the naming convention "latest_instances_<user>.html" is created at the script execution location. The file contains a listing of retrieved instances.
   8. The macro and script execution ends with the printing of the line "EXITING IRM_REST_RECREATE_INSTANCES" to the execution log.
      
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2017
*/

%macro irm_rest_recreate_instances(
      host                 =
   ,  port                 = 7980
   ,  username             =
   ,  password             =
   ,  server               = 
   ,  debug                = false
   ,  logOptions           =
   ,  final_fa_id          =
   ,  pollInterval         = 
   ,  maxWait              =
   ,  script_loc           =
   ,  pa_location          =
);
                                   
   %put ENTERING &sysmacroname.; 

   %local mySrvTicket;
   %let mySrvTicket=;

   /* Set the required log options */
   %if(%length(&logOptions.)) %then
   options &logOptions.;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   %if(%length(&port.) = 0) %then
      %let port = 7980;

   %if(%rsk_dsexist(instances_info)) %then %do;
      proc datasets lib=work nolist;
         delete instances_info;
      quit;
   %end;

   /* Retrieve information about existing instances in the instances_info dataset */
   %irm_rest_get_instances(
         host         = &host. 
      ,  port         = &port. 
      ,  username     = &username. 
      ,  password     = &password.
      ,  outds        = instances_info 
      ,  outVarTicket = &mySrvTicket.
   );
         
   %let user=%trim(%left(&username.));

   %if(%rsk_dsexist(instances_info)) %then %do;
      /* Sort the dataset by instance name. */
      proc sort data=instances_info;
         by name;
      run;

      /* Print existing instances info to a html file at the script location. */
      options nocenter;
      ods listing close;
      ods html style=htmlBlue path="&script_loc./" body="existing_instances_&user..html";
      title "Listing of Existing Instances (before script execution) created by/shared with &user. ";

      proc print data=instances_info;
      run;
       
      /* Close the ODS destination */
      ods html close;   

      /* Update federatedAreaId to the highest fa id */
      data instances_info_fa_id_updt;
         set instances_info;
         federatedAreaId = "&final_fa_id.";
      run;

      /* Filter out instances not created by the user running the script but shared by another user. 
         If retained, original instance will not be deleted but duplicate instance by the current user will be created. 
         User running the script could be an internal user (@saspw), local user on the machine or an external domain account. 
         If internal, the username contains a @saspw suffix. However, instances created by an internal user do not contain the
         @saspw suffix in created_by column in jobflow_instance. But the username with @saspw suffix is required for REST API authorization. 
         Same situation arises for a domain user such as user1@<domain>. The username is stored as user1 but user1@<domain> is required for
         authentication. The possibilities are listed below:
            1. irmuser1 (a regular local user)  DB stores as irmuser1
            2. irmuser1@saspw (an internal user)  DB stores as irmuser1
            3. irmuser1@test (a regular local user)  DB stores as irmuser1@test
            4. irmuser1@test@saspw (an internal user)  DB stores as irmuser1@test
            5. irmuser1@carynt (a domain user)  DB stores as irmuser1
            6. irmuser1@test@carynt (a domain user)  DB stores as irmuser1@test

       The logic used here is to strip everything from and including the last @. This will satisfy all scenarios except for #3.
     */ 

      %let srch_chr=@; /* search character */
      %if %index(&user.,&srch_chr.) gt 0 %then %do; /* contains @ */
         %let user_int=%substr(&user.,1,%sysfunc(findc(&user.,&srch_chr.,B))-1); /* strip out everything from and including the last @ */
         /* Retain instances created by the user. */
         data instances_info_fltrd;
            set instances_info(where=(upcase(createdBy)=upcase("&user_int.")));
         run;
      %end;
      %else %do; /* not an internal user */
         /* Retain instances created by the user. */
         data instances_info_fltrd;
            set instances_info(where=(upcase(createdBy)=upcase("&user.")));
         run;
      %end;

      /* Subset list of instances based on status codes and modified input. */
      /***********************************************************
         Status Codes:  
         0 = UNKNOWN,            -Default status
         1 = INVALID,            -Process or Task is not valid (based on BPMN and own SII logic)
         2 = RUNNING,            -Process or task is running              
         3 = COMPLETED_ERROR,    -Completed with error           
         4 = COMPLETED_SUCCESS,  -Successfully completed              
         5 = CANCELLED,          -Cancelled               
         6 = PUBLISHED,          -Published
         7 = OUT_OF_DATE,        -Out of date               
         8 = PUBLISHING,         -Publishing
         9 = PENDING             -Generic pending state
       ***********************************************************/
       
      data published_instances_info modified_instances_info other_instances_info;
         set instances_info_fltrd; /* already filtered for the user */
         if (status in (6,8)) then do; /* PUBLISHED and PUBLISHING */
            output published_instances_info;
         end;
         else do;
            if (status in (4)) then do; /* COMPLETED_SUCCESS */
               if (upcase(inputDataModified)='TRUE') then do;
                  output modified_instances_info;
               end;
               else do;
                 output other_instances_info;
               end;
            end; 
         end;
      run;

      %if %rsk_attrn(modified_instances_info, NOBS) eq 0 %then %do;        
         %put NO INSTANCES WITH MODIFIED INPUTS FOR RE-CREATION.;
      %end;
      %else %do;
         /* Call macro to re-create instances with modified inputs. */
         %irm_rest_run_mod_instance(  
               input_instance_info  = modified_instances_info 
            ,  pa_loc               = &pa_location.
            ,  mhost                = &host.
            ,  mport                = &port.
            ,  musername            = &username.
            ,  mpassword            = &password.
            ,  mserver              = &server.
            ,  mdebug               = &debug.
            ,  mlogOptions          = &logOptions.
            ,  mfinal_fa_id         = &final_fa_id.
            ,  mpollInterval        = &pollInterval.
            ,  mmaxWait             = &maxWait.
         );
      %end;
    
      /* Fetch the number of flow instances that have to be recreated */
      %let numInstances = %rsk_attrn(other_instances_info, NOBS);
      
      %if &numInstances. eq 0 %then %do;        
         %put NO UN-PUBLISHED and NON-MODIFIED INSTANCES FOR RE-CREATION;
      %end;
      %else %do;
         /* First delete the instance and then re-create it */
         %do i=1 %to &numInstances;
            /* Create macro variables for instance attributes. */
            data _null_;
               set other_instances_info(firstobs=&i. obs=&i.);
               call symputx('instanceKey',key);
               call symputx('configset',configSetId);
               call symputx('entityId',entityId);
               call symputx('entityRole',entityRoleKey);
               call symputx('basedate',baseDate);
               call symputx('category',categoryId);
               call symputx('bpmnname',jobFlowDefinitionFile);
               call symputx('desc',description);
               call symputx('instanceName',name);
               call symputx('shared', visibility);
            run;

            %put Instance Visibility = &shared.;

            /* Delete the instance */
            %irm_rest_manage_instance(    
                  instanceKey = &instanceKey.
               ,  host        = &host.
               ,  LogonPort   =
               ,  server      = &server.
               ,  port        = &port.
               ,  username    = &username.
               ,  password    = &password.
               ,  outds       = 
               ,  action      = delete
               ,  comment     =
               ,  debug       = &debug.
               ,  logOptions  = &logOptions.
            )

            /* Re-create the instance */
            %let NewInstanceKey =;

            %irm_rest_create_jobflow_instance(    
                  instance_name   = %bquote(&instanceName.)
                , description     = &desc.
                , entityID        = &entityId.
                , entityRoleKey   = &entityRole.
                , category        = &category. 
                , jobflowFile     = &bpmnname.
                , baseDate        = &basedate.
                , configSetId     = &configset.
                , federatedAreaID = &final_fa_id.
                , host            = &host.
                , port            = &port.
                , username        = &username.
                , password        = &password.
                , outvar          = NewInstanceKey
                , outVarTicket    = mySrvTicket
                , logOptions      = &logOptions.
            );

            %put NewInstanceKey = &NewInstanceKey.;

            %if (%sysevalf(%superq(NewInstanceKey)=,boolean)) or ("%substr(&NewInstanceKey,1,2)" = "<?") %then
                  %put ERROR:  There was a problem creating an instance for %str(&instanceName.);
            %else %do;
               %let myticket=;
               option nonotes;
               /* Wait for the running instance to complete execution. */
               %irm_rest_wait_for_job(   
                     instanceKey  = &NewInstanceKey.
                  ,  host         = &host.
                  ,  port         = &port.
                  ,  username     = &username.
                  ,  password     = &password.
                  ,  pollInterval = &pollInt.
                  ,  maxWait      = &maxWait.
                  ,  outds        = jobflow_status
                  ,  outVarTicket = myticket
                  ,  debug        = &debug.
                  ,  logOptions   = &logOptions.
               );
               
               %if (%upcase(&shared.) eq PUBLIC) %then %do;
                  /* Share the instance. */
                  %irm_rest_manage_instance(    
                        instanceKey = &NewInstanceKey.
                     ,  host        = &host.
                     ,  LogonPort   =
                     ,  server      = &server.
                     ,  port        = &port.
                     ,  username    = &username.
                     ,  password    = &password.
                     ,  outds       = shared_status
                     ,  action      = share
                     ,  comment     =
                     ,  debug       = &debug.
                     ,  logOptions  = &logOptions.
                  )
               %end; /* End of %if (%upcase(&shared) eq PUBLIC) %then %do; */ 
            %end; /* End of %if (%sysevalf(%superq(NewInstanceKey)=,boolean)) */
         %end; /* End of do loop %do i=1 %to &numInstances; */
      %end; /* End of else block for %if &numInstances. eq 0 %then %do; */
   %end; /* End of %if(%rsk_dsexist(instances_info)) %then %do; */
   
   /* Obtain a final list of all instances in the new_instances_info dataset. */
   %let ticket=;
   %irm_rest_get_instances(
         host         = &host. 
      ,  port         = &port. 
      ,  username     = &username. 
      ,  password     = &password.
      ,  outds        = new_instances_info 
      ,  outVarTicket = &ticket.
   );
   
   /* Print latest instance info to a file at the script location. */
   %if(%rsk_dsexist(new_instances_info)) %then %do;
       /* Sort the dataset by instance name. */
       proc sort data=new_instances_info;
         by name;
       run;

       /* Print existing instance info to a html file at the script location. */
       options nocenter;
       ods listing close;
       ods html style=htmlBlue path="&script_loc./" body="latest_instances_&user..html";
       title "Listing of Latest Instances (after script execution) created by/shared with &user.";

       proc print data=new_instances_info;
       run;
       
       /* Close the ODS destination */
       ods html close;  
   %end;

   %put EXITING &sysmacroname.;

%mend;
