/*
 Copyright (C) 2017 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief  This macro will recreate the jobflow instances wchich have modified inputs. 
           Based on the input instance information table(which only contains those instances that have been input modified), 
           thie macro exports the modified input files into excel files, deletes the existing instance, recreates the instance and reuploads 
           the excel files. 
   
   <b> Identified Inputs </b>
   
   \param[in] input_instance_info Input table which contains the instances information
   \param[in] PA_LOC              Persistent area location
   \param[in] mhost               Host url, including the protocol
   \param[in] mmaxWait            Maximum waiting time for waiting the jobflow instance to complete. 
   \param[in] mserver             Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param[in] mport               Server port (if blank, default port 80 will be used). Default: 7980
   \param[in] musername           credentials
   \param[in] mpassword           credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param[in] mfinal_fa_id        Name of the final federated ID
   \param[in] mpollInterval       The interval time for checking if the job flow instance is completed.
   \param[in] mdebug              True/False. If True, debugging informations are printed to the log (Default: false)
   \param[in] mlogOptions         Logging options (i.e. mprint mlogic symbolgen ...)
      
   \details 
           1. Loop through every instance in the input instance information table(which only contains those instances that have been input modified). By going into PA location, 
           identify the data sets that have been modified. Based on the library mapping table(user prepare), find out the library this table belongs to. The process will abort if the 
           lib_mapping table do not use the correct delimeters for path. And it will abort if no library could be found for a modified file. And it will abort if data key could not be retrived from 
           a given data set in the recreated instance. 
           
           2. export all the data sets to excel files.
           
           3. delete the instance and recreate a new one. 
           
           4. Loop through every excel file, identify the data key for that table in the instance. upload the excel file to the instance which trigger the rerun. If upload failed, Error will
           output to the log. 
     
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2017
*/
%macro irm_rest_run_mod_instance(     input_instance_info= 
                                       , PA_LOC=
                                       , mhost =
                                       , mport = 7980
                                       , musername =
                                       , mpassword =
                                       , mserver = 
                                       , mdebug = false
                                       , mlogOptions =
                                       , mfinal_fa_id =
                                       , mpollInterval = 
                                       , mmaxWait      =

                               );
                               


   /* Fetch the number of flow instances that have to be recreated */
   %let numInstances = %rsk_attrn(&input_instance_info., NOBS);
   
   /* in order to read in the library mapping table*/
   libname lib_map "&PA_LOC.";
   
   /* first checking the lib_mapping table, if it has the correct delimeter*/
   %if ((&SYSSCP eq WIN) or (&SYSSCP eq DNTHOST))%then %do;
      %let NODEL=/; 
   %end;
   %else %do; 
      %let NODEL=\;
   %end;
   
   data temp;
      set lib_map.lib_mapping;
      x = index(LA_PATH, "&NODEL");
      if x>0; output;
   run;
   
   %if %rsk_attrn(work.temp, NLOBS) > 0 %then %do;
      %put "ERROR: Library mapping table does have the correct delimeter. Windows \ and Linux /.";
      %rsk_terminate();
   %end;

   %do i=1 %to &numInstances;
      /* Create macro variables for instance attributes. */
      data _null_;
            set &input_instance_info.(firstobs=&i. obs=&i.);
            call symputx('instanceKey',key);
            call symputx('configset',configSetId);
            call symputx('entityId',entityId);
            call symputx('entityRole',entityRoleKey);
            call symputx('basedate',baseDate);
            call symputx('category',categoryId);
            call symputx('bpmnname',jobFlowDefinitionFile);
            call symputx('desc',description);
            call symputx('instanceName',name);
            call symputx('shared', visibility);
      run;

      /*before the old instance gets deleted, search all the files that needs to be uploaded*/
      %local FILE_PATH DEL;
      %if ((&SYSSCP eq WIN) or (&SYSSCP eq DNTHOST))%then %do;
         %let FILE_PATH=&PA_LOC.\data\&instanceKey.\landing_area;
         %let DEL=\; 
      %end;
      %else %do; 
         %let FILE_PATH=&PA_LOC./data/&instanceKey./landing_area;
         %let DEL=/;
      %end;

      %read_flnm(fileloc=&FILE_PATH.,
                       fileextension=sas7bdat,
                       flg_subdirs=Y, 
                       flg_unique=N, 
                       wildcard=, 
                       wc_type=startwith,
                       exception_file=system_option.sas7bdat, 
                       result=data_set_modified);
     
      /* library mapping, only output those files that have LIARARY resolved */ 
      proc sql noprint;
         create table work.lib_rev_file as
         select a.*, b.LIBRARY
         from work.data_set_modified as a left join lib_map.lib_mapping as b
         on upcase(trim(a.path_rel))=upcase(trim(b.LA_PATH));
      quit;

      /* if data set lib_rev_file has LIBRARY not resolved, abort*/
      data work.lib_rev_file_nolib;
          set work.lib_rev_file;
          if LIBRARY eq "" then 
          output; 
      run;
      
      %if %rsk_attrn(work.lib_rev_file_nolib, NLOBS) > 0 %then %do;
         %put 'ERROR: Some modified data from instance '"&instanceKey."' could not find the corresponding library from lib_mapping table.';
         %rsk_terminate();
      %end;
      
      data work.lib_rev_file;
          set work.lib_rev_file;
          file_name=cats(LIBRARY, '.', name_root);
          %if ((&SYSSCP eq WIN) or (&SYSSCP eq DNTHOST))%then %do;
             file_path=cats(path,'\',name_full);
          %end;
          %else %do;
             file_path=cats(path,'/',name_full);
          %end;
      run;
      
      /* start the export the modified input files to excel file */
      %let nobs =%rsk_attrn(work.lib_rev_file, NLOBS);
      %let NDS=1;
      %do %while(&NDS. <= &nobs.);

        data _null_;
           set work.lib_rev_file(FIRSTOBS=&NDS. OBS=&NDS.);
           call symput('FILE_NM', file_name);
           call symput('FILE_PATH', file_path);
           call symput('Lib_nm', upcase(Library));
           call symput('Lib_Path', trim(path));
           call symput('Root_NM', trim(name_root));
           call symput('up_root_nm',upcase(trim(name_root)));
        run;

        libname &Lib_nm. "&Lib_Path.";       

        %local out_loc;
        %let out_loc=&PA_LOC.&DEL.backup&DEL.&instanceKey.&DEL.;
        %put &out_loc.;
        %rsk_mkdirs_and_verify(&out_loc.);
        
        PROC EXPORT DATA=&FILE_NM.
                    OUTFILE="&out_loc.&Root_NM..xlsx"
                    DBMS=XLSX REPLACE;
        RUN;

        libname &Lib_nm. clear;
        %let NDS=%eval(&NDS+1);

      %end;

      /* First delete the instance and then re-create it */
      %irm_rest_manage_instance(    instanceKey = &instanceKey.
                                 , host = &mhost.
                                 , LogonPort =
                                 , server = SASIRMServer
                                 , port = &mport.
                                 , username = &musername.
                                 , password = &mpassword.
                                 , outds = 
                                 , action = delete
                                 , comment =
                                 , debug = &mdebug.
                                 , logOptions = &mlogOptions.
                                 );

      /* Re-create the instance */
      %let NewInstanceKey =;
      %let myCreateTicket =;

      %irm_rest_create_jobflow_instance(    instance_name   = %bquote(&instanceName.)
                                           , description     = &desc.
                                           , entityID        = &entityId.
                                           , entityRoleKey   = &entityRole.
                                           , category        = &category. 
                                           , jobflowFile     = &bpmnname.
                                           , baseDate        = &basedate.
                                           , configSetId     = &configset.
                                           , federatedAreaID = &mfinal_fa_id.
                                           , host            = &mhost.
                                           , port            = &mport.
                                           , username        = &musername.
                                           , password        = &mpassword.
                                           , outvar          = NewInstanceKey
                                           , outVarTicket    = myCreateTicket
                                           , logOptions      = &mlogOptions.
                                           );

      %put NewInstanceKey = &NewInstanceKey.;

      %if (%sysevalf(%superq(NewInstanceKey)=,boolean)) or ("%substr(&NewInstanceKey,1,2)" = "<?") %then
         %put ERROR:  There was a problem creating an instance for %str(&instanceName.);
      %else %do;
         %let myticket=;
         option nonotes;

          /* Wait for the running instance to complete execution. */
         %irm_rest_wait_for_job(   instanceKey  = &NewInstanceKey.
                                 , host         = &mhost.
                                 , port         = &mport.
                                 , username     = &musername.
                                 , password     = &mpassword.
                                 , pollInterval = &mpollInterval.
                                 , maxWait      = &mmaxWait.
                                 , outds        = jobflow_status
                                 , outVarTicket = myticket
                                 , debug        = &mdebug.
                                 , logOptions   = &mlogOptions.
                                 );

      %end; /* End of %if (%sysevalf(%superq(NewInstanceKey)=,boolean)) */

      /* Get the dataKey needed to upload the input file */
      %irm_rest_get_table_map(instanceKey = &NewInstanceKey
                                , host = &mhost.
                                , port = &mport.
                                , username = &musername.
                                , password = &mpassword.
                                , outds = jobflow_table_map
                               );

      /* start the upload the modified input files */
      %let nobs =%rsk_attrn(work.lib_rev_file, NLOBS);
      %let NDS=1;
      %do %while(&NDS. <= &nobs.);

        data _null_;
           set work.lib_rev_file(FIRSTOBS=&NDS. OBS=&NDS.);
           call symput('FILE_NM', file_name);
           call symput('FILE_PATH', file_path);
           call symput('Lib_nm', upcase(Library));
           call symput('Lib_Path', trim(path));
           call symput('Root_NM', trim(name_root));
           call symput('up_root_nm',upcase(trim(name_root)));
        run;

        %let dataKey1=;

         data _null_;
            set jobflow_table_map(where=(upcase(libname) = "&Lib_nm." and upcase(TableName) = "&up_root_nm."));
            call symput("dataKey1", strip(dataKey));
         run;

         %if &dataKey1. ne  %then %do;
            /* Upload the input file */
            %irm_rest_upload_file(host = &mhost.
                               , port = &mport.
                               , username = &musername.
                               , password = &mpassword.
                               , topInstanceKey = &NewInstanceKey.
                               , dataKey = &dataKey1.
                               , dirPath = &out_loc.
                               , fileName = &Root_NM..xlsx
                               );

            %irm_rest_wait_for_job(instanceKey  = &NewInstanceKey.
                                 , host         = &mhost.
                                 , port         = &mport.
                                 , username     = &musername.
                                 , password     = &mpassword.
                                 , pollInterval = &mpollInterval.
                                 , maxWait      = &mmaxWait.
                                 , outds        = jobflow_status
                                 , outVarTicket = myticket
                                 , debug        = &mdebug.
                                 , logOptions   = &mlogOptions.
                                 ); 
                                 
            %if (%upcase(&shared.) eq PUBLIC) %then %do;
               /* Share the instance. */                                 
               %irm_rest_manage_instance(    
                                    instanceKey = &NewInstanceKey.
                                 ,  host        = &mhost.
                                 ,  LogonPort   =
                                 ,  server      = SASIRMServer
                                 ,  port        = &mport.
                                 ,  username    = &musername.
                                 ,  password    = &mpassword.
                                 ,  outds       = mod_shared_status
                                 ,  action      = share
                                 ,  comment     =
                                 ,  debug       = &mdebug.
                                 ,  logOptions  = &mlogOptions.
                              );
            %end; /* End of %if (%upcase(&shared.) eq PUBLIC) */     
         %end;  
         %else %do;
            %put: 'ERROR: data set '"&Lib_nm." "&up_root_nm."' could not find the datakey for the instance';
         %end;
         
         %let NDS=%eval(&NDS+1);


      %end; /* End of NDS looping */
      PROC Delete data =work.lib_rev_file ; 
      PROC Delete data =work.data_set_modified ;
      PROC Delete data =work.jobflow_table_map ;
      
   %end; /* End of do loop %do i=1 %to &numInstances; */
 

%put Exiting &sysmacroname.;
                         
%mend;
