/*
 Copyright (C) 2017 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Upload a file inside a jobflow instance
   
   \param [in] host Host url, including the protocol
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] port Server port (if blank, default port 80 will be used). Default: 7980
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] authMethod: Authentication method (accepted values: TOKEN/SSO). Used when Username/Password credentials are left blank to determine the authentication method. (Default: TOKEN). 
   \param [in] topInstanceKey Instance key of the top flow
   \param [in] dataKey Data object unique key
   \param [in] dirpath Full path to the file folder
   \param [in] filename name of the file to upload
   \param [in] comment Comment attached to the file
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
      
   \details 
   This macro sends a POST request to <b><i>\<host\>:\<port\>/SASIRMServer/rest/jobflow/instance/&topInstanceKey./import/data/&dataKey.</i></b> in order to upload a file into the specified instance.
   Use macro \link irm_rest_get_table_map.sas \endlink in order to obtain the keys required for uploading the file.      
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2017
*/
%macro irm_rest_upload_file(host =
                            , LogonPort =
                            , server = SASIRMServer
                            , port = 7980
                            , username =
                            , password =
                            , authMethod = token /* token/sso */
                            , topInstanceKey =
                            , dataKey     =
                            , dirPath     =
                            , fileName    =
                            , comment     = %str(Input was modified by user)
                            , debug       = true
                            , logOptions =
                            );
   

   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;

   /* Check if username and password have been provided. */
   %local without_credentials_flg;
   %let without_credentials_flg = N;
   %if(%sysevalf(%superq(username)=, boolean) and %sysevalf(%superq(password)=, boolean)) %then
      %let without_credentials_flg = Y;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   %if(%length(&port.) = 0) %then
      %let port = 80;
   
   %let url = %str(&host.:&port./&server./rest/jobflow/instance/&topInstanceKey./import/data/&dataKey./);

   /* Get a Service ticket in the following cases:
      - Credentials have been provided => authenticate with the credentials
      - The system is configured to use Single Sign-On (authMethod = sso) => use proc http AUTO_NEGOTIATE option
   */
   %if(&without_credentials_flg. = N or %upcase(&authMethod.) = SSO) %then %do;
      /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
      option nomlogic nosymbolgen;
      /* Get the Service Ticket */
      %local ticket;
      %irm_rest_get_ticket(url = &url.
                           , LogonPort = &LogonPort.
                           , server = &server.
                           , username = &username.
                           , password = &password.
                           , outVarTicket = ticket
                           , debug = &debug.
                           , logOptions = &oldLogOptions.
                           );
   %end;
   
   filename hin  temp;
   filename hout temp;
   filename out  temp;
   /* HTTP protocol requires CRLF for multipart streaming */
   filename in   temp termstr = crlf;
   filename in2  temp termstr = crlf;

   data _null_;
      file hin;
      put 'accept: */*';
   run;

   /* Create the payload (multi-part document) */
   data _null_;
      file in;
      put "-----------------------------26704164382695";
      put "Content-Disposition: form-data; name=""file""; filename=""&FILENAME.""";
      put "Content-Type: application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
      put;
   run;

   data _null_;
      file in2;
      put;
      put "-----------------------------26704164382695";
      put "Content-Disposition: form-data; name=""_charset_""";
      put;
      put "UTF-8";
      put "-----------------------------26704164382695";
      put "Content-Disposition: form-data; name=""file-data""";
      put;
      put;
      put "-----------------------------26704164382695";
      put "Content-Disposition: form-data; name=""comment""";
      put;
      put "&comment";
      put "-----------------------------26704164382695--";
   run;

  %irm_file_append(file = &dirPath./&fileName.
                   , toFile = in
                   );
  %irm_file_append(file = in2
                   , toFile = in
                   );

   /* Send the request */
   proc http
         headerin  = hin 
         headerout = hout
         in        = in
         out       = out
         method    = 'post'
         ct        = 'multipart/form-data; boundary=---------------------------26704164382695'
         %if(&without_credentials_flg. = Y and %upcase(&authMethod.) = TOKEN) %then %do;
            /* No credentials -> use Token authentication (this is only allowed if running the code from an authenticated SAS Session) */
            url = "%str(&url.)"
            http_tokenauth
         %end;
         %else %do;
            /* Use the ticket */
            url  = "&url?ticket=&ticket." 
         %end;
         /* Clear connection cache */
         clear_cache
         ;
   run;

   /* Collect the response from the server and print it to the log */
   data _null_;
      infile out;
      input;
      put 'Message from server:';
      put _infile_;
      if index(_infile_,"500") > 0 then do;
         put 'ERROR: the upload failed for file '"&fileName."' in instance '"&topInstanceKey.";
         call execute('%rsk_terminate()');
      end;
   run;
   
   
 
   filename hin  clear;
   filename hout clear;
   filename out  clear;
   filename in   clear;
   filename in2  clear;
   
%mend;
