/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   
   \brief   Wait until the job execution is complete or the maximum waiting time has expired.
   
   \param [in] instanceKey Unique identifier of the instance
   \param [in] LogonPort (Optional) port where the /SASLogon web application is listening. If blank, it is assumed that the /SASLogon is on the same port as /SASIRMServer (Default: blank)
   \param [in] host Host url, including the protocol
   \param [in] server Name of the Web Application Server that provides the REST service (Default: SASIRMServer)
   \param [in] port Server port (if blank, default port 80 will be used). Default: 7980
   \param [in] username Username credentials
   \param [in] password Password credentials: it can be plain text or SAS-Encoded (it will be masked during execution). 
   \param [in] outds Name of the output table (Default: jobflow_status)
   \param [in] outVarTicket Name of the output macro variable which will contain the Service Ticket (Default: ticket)
   \param [in] pollInterval Number of seconds to wait in between consecutive requests to the server (Default: 5)
   \param [in] maxWait Maximum amount of seconds the macro will wait before giving up (Default: 3600 -> 1 hour)
   \param [in] debug True/False. If True, debugging informations are printed to the log (Default: false)
   \param [in] logOptions Logging options (i.e. mprint mlogic symbolgen ...)
      
   \details 
   This macro sends a GET request to <b><i><host>:<port>/SASIRMServer/rest/jobflow/status/&instanceKey.</i></b> and checks the status of the jobflow:
   - If the jobflow status is "Running" then it will wait the specified number of seconds (<i>pollInterval</i>) before checking again.
   - Irrespectively of the jobflow status, the macro will stop checking after the specified number of seconds (<i>maxWait</i>)
   Final status of the jobflow upon macro completion are collected in the output table. \n
   \n
   See \link irm_rest_get_request.sas \endlink for details about how to send GET requests and parse the response.
      
   <b>Example:</b>
   
   1) Set up the environment (set SASAUTOS and required LUA libraries)
   \code
      %let source_path = <Path to the root folder of the Federated Content Area (root folder, excluding the Federated Content folder)>;
      %let fa_id = <Name of the Federated Area Content folder>;
      %include "&source_path./&fa_id./source/sas/ucmacros/irm_setup.sas";
      %irm_setup(source_path = &source_path.
                , fa_id = &fa_id.
                );
   \endcode

   2) Check the jobflow status every 5 seconds until the jobflow has completed. Stop waiting after 1 hour (3600 secs).
   Final jobflow status is returned in the output table WORK.jobflow_status
   \code
      %let ticket =;
      %irm_rest_wait_for_job(instanceKey = <instance key>
                             , host = <host>
                             , port = <port>
                             , username = <userid>
                             , password = <pwd>
                             , pollInterval = 5
                             , maxWait = 3600
                             , outds = jobflow_status
                             , outVarTicket = ticket
                             );
   \endcode
         
   <b>Sample output:</b>

   | status     | type    | key         | status_code  |
   |------------|---------|-------------|--------------|
   | Successful | JOBFLOW | 1596921364  | 4            |
   | Successful | SUBFLOW | 133338160   | 4            |
   | Successful | SUBFLOW | 1785155517  | 4            |
   | Successful | SUBFLOW | 484874770   | 4            |
   | Successful | SUBFLOW | 1813302815  | 4            |
   | Successful | SUBFLOW | 2052723883  | 4            |
   | Successful | SUBFLOW | 189427956   | 4            |
   | Successful | SUBFLOW | 1113517225  | 4            |
   | Successful | TASK    | 1581914008  | 4            |
   | Successful | TASK    | 1563079177  | 4            |
   | Successful | TASK    | 1025039774  | 4            |
   | Successful | TASK    | 1888046022  | 4            |
   
   \ingroup macroUtils
   
   \author  SAS Institute Inc.
   \date    2015
*/
%macro irm_rest_wait_for_job(instanceKey =
                             , host =
                             , LogonPort =
                             , server = SASIRMServer
                             , port = 7980
                             , username =
                             , password =
                             , outds = jobflow_status
                             , outVarTicket = ticket
                             , pollInterval = 5
                             , maxWait = 3600
                             , debug = false
                             , logOptions =
                             );

   /* Set the required log options */
   %if(%length(&logOptions.)) %then
      options &logOptions.;
   ;

   /* Get the current value of mlogic and symbolgen options */
   %local oldLogOptions;
   %let oldLogOptions = %sysfunc(getoption(mlogic)) %sysfunc(getoption(symbolgen));

   %if(%length(&port.) = 0) %then
      %let port = 80;
   
   %local done start_dttm;
   %let done = N;
   %let start_dttm = %sysfunc(datetime());
   %do %while (&done. = N);
      proc datasets library = work nolist nodetails;
         delete _tmp_jobflow_status_;
      quit;
      
      /* Temporary disable mlogic and symbolgen options to avoid printing of userid/pwd to the log */
      option nomlogic nosymbolgen;
      /* Send the REST request */
      %irm_rest_get_request(url = &host:&port./&server./rest/jobflow/status/&instanceKey.
                            , LogonPort = &LogonPort.
                            , server = &server.
                            , username = &username.
                            , password = &password.
                            , parser = irmRestJobStatus
                            , outds = _tmp_jobflow_status_
                            , outVarTicket = &outVarTicket.
                            , debug = &debug.
                            , logOptions = &oldLogOptions.
                            );

      /* If something goes wrong then exit the loop */
      %let done = Y;
      
      %if(%rsk_dsexist(_tmp_jobflow_status_)) %then %do;
         data &outds.;
            set _tmp_jobflow_status_;
            if(type = "JOBFLOW" and status eq "Running") then
               call symputx("done", "N", "L");
         run;
      %end;
      
      %let current_dttm = %sysfunc(datetime());
      %if(&done. eq N) %then %do;   
         %if (%sysevalf(&current_dttm. - &start_dttm. > &maxWait.)) %then %do;
            %put ------------------------------------------------------------------------------------;
            %put WARNING: Maximum waiting time has expired before the job completion. Exiting macro..;
            %put ------------------------------------------------------------------------------------;
            %let done = Y;
         %end;
         %else %do;
            %put -------------------------------------------------------;
            %put Job is still running, waiting &pollInterval. seconds...;
            %put -------------------------------------------------------;
            %let rc = %sysfunc(sleep(&pollInterval., 1));
         %end;
      %end;
      
   %end;
                         
%mend;
