/**********************************************************************;
*NAME:       read_flnm                                    ;

*PURPOSE:    read files name from a directory in Linux or windows (or string from a file) into a sas data set, a macro variable and output to a text file

*USAGE:       %read_flnm(fileloc=, fileextension=sas, wildcard=,wc_type=, 
                    exception=, result=file_names, outMcrName=, outfilefullname=);

*INPUT:       fileloc: the directory path where these files reside, or full path name of the input file.
              fileextension: extension of file(file extension) to be read in (e.g., if fileextension=sas, then only read in sas files.). N.A. for reading from file.
              wildcard: wildcard character
              wc_type: only accept four type: 'startwith', 'endwith', 'contains', 'exact'.
              exception: filename with extension (or strings) to be excluded from reading in. Specify full name, including file extension.
                flg_unique: flag, takes Y or N. indicate if result data sets only contain unique string. string is also unique in &outMcrName.
              result: the name of output data set with files (or interested string) name. Must not be empty
                    
*OUTPUT:     A data set (&outfilefullname) a macro var(&outMcrName) and a text file(&outfilefullname.txt) with files name (or interested string)

**********************************************************************/
%macro read_flnm(fileloc=, flg_subdirs=, flg_unique=, fileextension=, wildcard=,wc_type=, exception_file=, 
      exception_dir=, result=);

   /************************************************************************************/
   /*******************Error checking for input params *************/
   /**************************************************************************************/
   
   /*error checking for fileloc*/
   %if  %bquote(&fileloc.) EQ %bquote() %then %do;
      %put ERROR: Must specify the starting directory;
      %return;
   %end;
   /*error checking for flg_subdirs*/
   %if  %bquote(&flg_subdirs.) EQ %bquote() %then %do;
      %put ERROR: Must specify the subdirs flag;
      %return;
   %end;
   /*error checking for flg_unique*/
   %if  %bquote(&flg_unique.) EQ %bquote() %then %do;
      %put ERROR: Must specify the unique flag;
      %return;
   %end;
   /*error checking for result*/
   %if  %bquote(&result.) EQ %bquote() %then %do;
      %put ERROR: Must specify the name for data set contains file names;
      %return;
   %end;

/*
   %let exception_file = %sysfunc(compress(&exception_file, "'"));
   %put unquoted exception file = &exception_file;
*/   
   /************************************************************************************/
   /*******************convert input parameter to upper case*************/
   /**************************************************************************************/
   %let  fileextension=%upcase(&fileextension.);
   %let  wildcard=%upcase(&wildcard.);
   %let  wc_type=%upcase(&wc_type.);
   %let  exception_file=%upcase(&exception_file.);
   %let  exception_dir=%upcase(&exception_dir.);
   %let  flg_unique=%upcase(&flg_unique.);
   %let  flg_subdirs=%upcase(&flg_subdirs.);
   

   /************************************************************************************/
   /*******************Error checking for wildcard and wc_type *************/
   /**************************************************************************************/
   /*error checking for wildcard*/
   %if  %bquote(&wildcard.) NE %bquote() and %bquote(&wc_type.) EQ %bquote() %then %do;
      %put ERROR: Must specify wildcard type(wc_type) since wildcard is specified;
      %return;
   %end;
   /*error checking for wc_type*/
   %if  %bquote(&wildcard.) NE %bquote() and %bquote(&wc_type.) NE %bquote() %then %do;
       %if %sysfunc(findw(&wc_type, STARTWITH, ' ', %bquote( I E))) EQ 0 and 
       %sysfunc(findw(&wc_type, ENDWITH, ' ', %bquote( I E))) EQ 0 and
       %sysfunc(findw(&wc_type, CONTAINS,' ', %bquote( I E))) EQ 0  and 
       %sysfunc(findw(&wc_type, EXACT, ' ', %bquote( I E))) EQ 0 %then %do;  
           %put Error: wc_type only can be either of startwith, endwith, contains, exact;
           %return;
       %end;
   %end;
   /*if wildcard is not specified, then empty wc_type*/
   %if  %bquote(&wildcard) EQ %bquote() %then %do;
      %let  wc_type=;
   %end;
   
   /*construct file extension filter, which is condition that select out NON-interested file*/
   %if  %bquote(&fileextension) NE %bquote() %then %do;
      %let fileextension_txt=%bquote(upcase(ext)~="&fileextension");
   %end;
   %else %do;
      %let fileextension_txt=%bquote(ext EQ '?');  /*extension equals to a question mark, which will never happen, so it will no filter out any obs*/
   %end;
   
   /************************************************************************************/
   /*******************start to grab insterested file*******************/
   /**************************************************************************************/
   /*    construct filter text(to be filtered out) according to input parameters*/
   %if %bquote(&wildcard.) NE %bquote() %then %do;
      %if %bquote(&wc_type.) EQ %bquote(STARTWITH) %then %do;
         %let filter_txt=%bquote(find(name_full, "&wildcard", 'i') NE 1);
      %end;
      %else %if %bquote(&wc_type.) EQ %bquote(ENDWITH) %then %do;
         %let filter_txt=%bquote(find(name_full, "&wildcard",'i', -100000) + LENGTH("&wildcard") -1 NE LENGTH(file_nm));
      %end;
      %else %if %bquote(&wc_type.) EQ %bquote(CONTAINS) %then %do;
         %let filter_txt=%bquote(find(name_full, "&wildcard",'i') EQ 0);
      %end;
      %else %if %bquote(&wc_type.) EQ %bquote(EXACT) %then %do;
         %let filter_txt=%bquote(findw(name_full, "&wildcard", ' ','Ei') EQ 0);
      %end;
      %else %do;
         %let filter_txt=;
      %end;
   %end;
   %else %do;
      %let filter_txt=;
   %end;

   /*use filename unnamed pipe to grab all file and subfolder names in &fileloc*/
   %local del;
   %if ((&SYSSCP eq WIN) or (&SYSSCP eq DNTHOST))%then %do;
      filename pp_files pipe "dir /b /s /a-d &fileloc";
   %end;
   %else %do; 
      filename pp_files pipe "find &fileloc";
   %end;
  
      
   data work.files1(drop=pos);
      infile pp_files lrecl=1024;
      input;
      length flg_dir $1 path $256 path_rel $128 name_full $128 name_root $128 extension $32;
      flg_dir='N';
      /* get fullpath */
      pos=findc(_infile_,'\/', 'B');
      path = substr(_INFILE_, 1, pos-1);
      put  path = ;

      /* get fullfile_name */
      name_full=substr(_INFILE_, pos+1);
      put name_full = ;

      /* get relative path */
      pos=length("&fileloc.");
      path_rel = substr(path, pos+1);
      put  path_rel = ;      

      /* get name_noext */
      pos=findc(name_full,'.', 'B');
      if pos GT 1 then do; /*if there is file extension*/
         name_root=substr(name_full, 1, pos-1);
         extension=substr(name_full, pos+1);
      end;
      if pos EQ 1 then do; /*if file start with a dot*/
         name_root=name_full;
         call missing(extension);
      end;
      if pos EQ 0 then do; /*if there is no file extension*/
         name_root=name_full;
         call missing(extension);
      end;
      put name_root = ;
      put extension = ;
      run;
   
   filename pp_files clear;

   data work.files;
      set work.files1;
   run;
   
   /* don't include subdirectories if flg_subdirs is not Y */
   %if  %upcase(&flg_subdirs.) NE Y %then %do;
      data work.files2(drop=pos) work.exclude2;
         set work.files;
         if upcase(path)=upcase("&fileloc") then do;
            output work.files2;   /*only grab files in &fileloc*/
         end;
         else do;
            output work.exclude2;
         end;
      run;
      data work.files;
         set work.files2;
      run;
   %end;

   /* filter out files that live in the exception directories list */
   %if  %bquote(&exception_dir.) NE %bquote() %then %do;
      %local path_rel;
      %global path_rel_macro;
      data work.files3 work.exclude3;
         set work.files;
         numwords = countw(path_rel, '\/');
         i=1;
         continue=1;
         do while (i<=numwords and continue);
            dir_part = scan(path_rel, i, '\/');
            put "dir_part = " dir_part;   
            if upcase(dir_part) in (%rsk_quote_list(list=&exception_dir)) then do;
               continue=0;
            end;

            i = i + 1;
         end;
         if continue then do;
            put "**** output record to files3: " path_rel;   
            output work.files3;
         end;
         else do;
            put "**** output record to exclude3: " path_rel;   
            output work.exclude3;
         end;      
      run;
      data work.files;
         set work.files3;
      run;   
   %end;

   /* filter out files in the exception files list */
   %if  %bquote(&exception_file.) NE %bquote() %then %do;
      data work.files4 work.exclude4;
         set work.files;
         if upcase(name_full) in (%rsk_quote_list(list=&exception_file)) then do;
            output work.exclude4;
         end;
         else output work.files4;
      run;
      data work.files;
         set work.files4;
      run;
   %end;

   /* filter out files not matching the extension */
   %if  %bquote(&fileextension.) NE %bquote() %then %do;
      %let extension_upcase = %upcase(&fileextension);
      data work.files5 work.exclude5;
         set work.files;
         if upcase(extension) = "&extension_upcase" then do;
            output work.files5;
         end;
         else output work.exclude5;
      run;
      data work.files;
         set work.files5;
      run;   
   %end;

   /* filter out files not matching the wildcard */
   %if  %bquote(&wildcard.) NE %bquote() %then %do;
      %let extension_upcase = %upcase(&fileextension);
      data work.files6 work.exclude6;
         set work.files;
         if %unquote(&filter_txt) then do;
            output work.exclude6;
         end;
         else output work.files6;
      run;
      data work.files;
         set work.files6;
      run;   
            
   %end;

   /*if indicate only unique strings in result data set*/
   /*
   */
   %if  (&flg_unique = 'Y') %then %do;
      proc sort data=work.files nodupkey;
         by path name_full;
      run;
   %end;

data work.&result.;
   set work.files;
run;   

   
%mend read_flnm;
