/*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   \brief The macro causes SAS job to exit with an error but does not terminate the stored process server or 
          workspace server (if running in that environment)..
      
   \details  Since most terminations put out a message prior to exiting, optional message info can be given as 
             input to put a msg to the log prior to exiting.  This message is in additional to the standard 
             message that a fatal error has occurred and that the code is exiting. 
      
   <b> Identified Inputs </b>  
   
   \param[in]  KEY                The message key 
  
   \param[in]  S1-S7              Message input
   
   The substitution parameters key,s1-s7 are all optional
   
    \ingroup Utility
   
    \author SAS Institute INC.
    \date 2015
    
 */
 
%macro rsk_terminate(KEY, s1, s2, s3, s4, s5, s6, s7);

   /* Optionally print an additional message prior to terminating. */
   %if (%bquote(&KEY.) ne ) %then %do;
      %rsk_print_msg(&KEY, &s1, &s2, &s3, &s4, &s5, &s6, &s7);
   %end;
   
   /* Print in the log that a fatal error has occurred and we are stopping execution. */
   %rsk_print_msg(rsk_system_fatal_error);

   /* Get the macro call stack to put in the log before aborting. */
   %local stack;
   %let stack=;
   %rsk_terminate_get_macro_stack(o_call_stack_var=stack);
   
   /* Log the macro call stack. */
   %local abort_msg;
   %let abort_msg = Abending at macro call stack;

   /* we are about to abort this sas session, help this error msg show up well in the log by disabling some logging options */   
   %put >>>--------------------------------------------------------------------------;
   %put >>> &abort_msg. ;
   %put >>> &stack. ;
   %put >>>--------------------------------------------------------------------------;
   %let SYSCC=9999; 
   /* Cause this SAS code to exit. */
   %abort cancel;

%mend rsk_terminate;


  /*
 Copyright (C) 2015 SAS Institute Inc. Cary, NC, USA
*/

/**
   \file
   \brief The macro returns the call stack trace of the macro calls made at this point. Only used within
          rsk_terminate, not use outside rsk_terminate.
      
   \details An important BAD side-effect of this macro, is that the SAS log is diverted to a temporary location 
            and can never be reliably returned to it's original location.
            So, only use this macro in cases where that is not a problem.
 
    \ingroup CommonAnalytics
   
    \author SAS Institute INC.
    \date 2015
    
 */
 
%macro rsk_terminate_get_macro_stack(O_CALL_STACK_VAR=);

   /* Save settings of options we care about  */

   %local savemlogic savenest savenotes savesource savesource2 savels;

   %let savemlogic=%sysfunc(getoption(MLOGIC));
   %let savenest=%sysfunc(getoption(MLOGICNEST));
   %let savenotes=%sysfunc(getoption(NOTES));
   %let savesource=%sysfunc(getoption(SOURCE));
   %let savesource2=%sysfunc(getoption(SOURCE2));
   %let savels=%sysfunc(getoption(LINESIZE, KEYWORD));

   options mlogic mlogicnest nonotes nosource nosource2 ls=256;

   /* Send the log to a temporary file in the work directory */

   filename temp temp lrecl=1024;

   proc printto log=temp new;
   run;

   /* Invoke a dummy macro so that the MLOGICNEST option has
      something to work with (has no effect on current macro) */

   %rsk_terminate_get_macro_stack_x;

   proc printto;
   run;

   /* Read the log to get the call stack information */

   data _NULL_;
      infile temp truncover lrecl=1024;
      input record $1024.;
      if (_N_ = 1) then do;
         re = prxparse("/MLOGIC\((.*)\.RSK_TERMINATE_GET_MACRO_STACK\./");
         retain re;
      end;
      if (prxmatch(re, record)) then do;
         length callstack $512;
         length dtstring $24;
         callstack = lowcase(prxposn(re, 1, record));
         callstack = TRANSTRN(callstack, ".", "-> ");
         call symput("&O_CALL_STACK_VAR", callstack);
         stop;
      end;
   run;

   /* Delete the temporary file */

   filename temp clear;

   /* Restore options we may have changed */

   options &savemlogic &savenest &savenotes &savesource &savesource2 &savels;

%mend rsk_terminate_get_macro_stack;

/* Dummy macro definition - do not delete this - needed by the main macro */
%macro rsk_terminate_get_macro_stack_x;
%mend rsk_terminate_get_macro_stack_x;
   
